import requests
import logging
import shutil
import time
import sys
import os

import dtlpy as dl

logger = logging.getLogger('dtlpy')
DOCKER_CONTAINER_NAME = 'dtlpy-development'


def spinning_cursor():
    while True:
        for cursor in ["(●             ●)",
                       "( ●           ● )",
                       "(  ●         ●  )",
                       "(   ●       ●   )",
                       "(    ●     ●    )",
                       "(     ●   ●     )",
                       "(      ● ●      )",
                       "(       ●       )",
                       "(      ● ●      )",
                       "(     ●   ●     )",
                       "(    ●     ●    )",
                       "(   ●       ●   )",
                       "(  ●         ●  )",
                       "( ●           ● )",
                       "(●             ●)",
                       ]:
            yield cursor


def copy_files_to_docker_container():
    from distutils.dir_util import copy_tree
    local_state_directory = os.path.dirname(dl.client_api.state_io.COOKIE)
    root_dir = os.path.dirname(local_state_directory)
    dtlpy_code_server_dir = os.path.join(dl.__path__[0], 'assets', 'code_server')
    local_code_server_dir = os.path.join(local_state_directory, 'code_server')
    copy_tree(src=dtlpy_code_server_dir, dst=local_code_server_dir)
    if not os.path.isdir(os.path.join(root_dir, '.vscode')):
        os.makedirs(os.path.join(root_dir, '.vscode'))
        shutil.copy(src=os.path.join(dl.__path__[0], 'assets', 'code_server', 'launch.json'),
                    dst=os.path.join(root_dir, '.vscode', 'launch.json'))
        shutil.copy(src=os.path.join(dl.__path__[0], 'assets', 'code_server', 'settings.json'),
                    dst=os.path.join(root_dir, '.vscode', 'settings.json'))


def get_container(client):
    try:
        container = client.containers.get(container_id=DOCKER_CONTAINER_NAME)
    except Exception:
        container = None
    return container


def start_session():
    try:
        import docker
    except (ImportError, ModuleNotFoundError):
        logger.error(
            'Import Error! Cant import "docker". Must install docker package for local development. run "pip install docker"')
        raise

    try:
        client = docker.from_env()
        client.ping()
    except Exception:
        raise RuntimeError(
            'Cannot communicate with docker daemon. Make sure everything is up and ready (docker client ping)'
        ) from None
    copy_files_to_docker_container()
    logger.info('Starting local development session')
    development = dl.client_api.state_io.get('development')
    port = development['port']
    docker_image = development['docker_image']
    hostname = f"http://localhost:{port}"
    logger.info(f'Starting docker image: {docker_image}, to {hostname}')

    need_to_create_container = True
    container = get_container(client=client)
    if container is not None:
        if not any([tag == docker_image for tag in container.image.tags]):
            # container with different image already running - delete it
            logger.info(f'Found a running image: {container.image.tags!r}, replacing to {docker_image!r}')
            container.stop()
            container.remove()
            need_to_create_container = True
        else:
            # container with SAME image already running - verify is running
            need_to_create_container = False

    if need_to_create_container:
        logger.info(f'Creating a container from docker image: {docker_image}')
        container = client.containers.run(docker_image,
                                          name=DOCKER_CONTAINER_NAME,
                                          tty=True,
                                          volumes={os.getcwd(): {'bind': "/tmp/app", 'mode': 'rw'}},
                                          ports={port: port},
                                          detach=True,
                                          command='/bin/bash')

    logger.info('Installing vscode on in the docker container')
    container = client.containers.get(container_id=DOCKER_CONTAINER_NAME)
    if container.status in "exited":
        container.start()
    elif container.status in "paused":
        container.unpause()

    resp = container.exec_run(
        cmd=f'/tmp/app/.dataloop/code_server/installation.sh {port}',
        detach=True,
        tty=True)
    spinner = spinning_cursor()
    while True:
        try:
            resp = requests.get(hostname, verify=False, timeout=0.5)
            success = resp.ok
        except Exception:
            success = False
        # write spinner
        sys.stdout.write(f'\r{next(spinner)}')
        sys.stdout.flush()
        time.sleep(0.1)
        sys.stdout.write('\b')

        # check response to break
        if success:
            logger.info(f'VScode server is up! {hostname}')
            break

    # open webbrowser to debug
    import webbrowser
    webbrowser.open(url=hostname, new=2, autoraise=True)


def pause_session():
    try:
        import docker
    except (ImportError, ModuleNotFoundError):
        logger.error(
            'Import Error! Cant import "docker". Must install docker package for local development. run "pip install docker"')
        raise

    try:
        client = docker.from_env()
        client.ping()
    except Exception:
        raise RuntimeError(
            'Cannot communicate with docker daemon. Make sure everything is up and ready (docker client ping)'
        ) from None

    container = get_container(client)
    if container is not None:
        container.pause()
    logger.info('Local development session paused')


def stop_session():
    try:
        import docker
    except (ImportError, ModuleNotFoundError):
        logger.error(
            'Import Error! Cant import "docker". Must install docker package for local development. run "pip install docker"')
        raise

    try:
        client = docker.from_env()
        client.ping()
    except Exception:
        raise RuntimeError(
            'Cannot communicate with docker daemon. Make sure everything is up and ready (docker client ping)'
        ) from None

    container = get_container(client=client)
    if container is not None:
        container.stop()
    logger.info('Local development session stopped')



# from torch.utils.tensorboard import SummaryWriter, FileWriter
# from tensorboard.compat.proto.event_pb2 import Event, SessionLog
# import time
#
#
# class DtlpyFileWrite(FileWriter):
#     def __init__(self, snapshot, *args, **kwargs):
#         self.snapshot = snapshot
#         super(DtlpyFileWrite, self).__init__(*args, **kwargs)
#
#     def add_event(self, event, step=None, walltime=None):
#         event: Event
#         event.wall_time = time.time() if walltime is None else walltime
#         if step is not None:
#             # Make sure step is converted from numpy or other formats
#             # since protobuf might not convert depending on version
#             event.step = int(step)
#         self.event_writer.add_event(event)
#         # print(event)
#         # print(event.step)
#         # print(event.SerializeToString())
#         dtlpy_records = list()
#         for v in event.summary.value:
#             dtlpy_records.append({'time': event.wall_time,
#                                   'step': event.step,
#                                   'tag': v.tag,
#                                   'value': v.simple_value,
#                                   'image': v.image.encoded_image_string,
#                                   'string': v.tensor.string_val,
#                                   'plugin_name': v.metadata.plugin_data.plugin_name})
#         print(dtlpy_records)
#         # self.snapshot.add_event(event)
#
#
# class DtlpySummaryWriter(SummaryWriter):
#     def __init__(self, snapshot, *args, **kwargs):
#         self.snapshot = snapshot
#         super(DtlpySummaryWriter, self).__init__(*args, **kwargs)
#
#     def _get_file_writer(self):
#         """Returns the default FileWriter instance. Recreates it if closed."""
#         print(self.snapshot)
#         if self.all_writers is None or self.file_writer is None:
#             self.file_writer = DtlpyFileWrite(snapshot=self.snapshot,
#                                               log_dir=self.log_dir,
#                                               max_queue=self.max_queue,
#                                               flush_secs=self.flush_secs,
#                                               filename_suffix=self.filename_suffix)
#             self.all_writers = {self.file_writer.get_logdir(): self.file_writer}
#             if self.purge_step is not None:
#                 most_recent_step = self.purge_step
#                 self.file_writer.add_event(
#                     Event(step=most_recent_step, file_version='brain.Event:2'))
#                 self.file_writer.add_event(
#                     Event(step=most_recent_step, session_log=SessionLog(status=SessionLog.START)))
#                 self.purge_step = None
#         return self.file_writer



def main():
    """
        Annotate a batch of images using a model and upload to platform
        :return:
        """
    import numpy as np
    from PIL import Image
    from keras.applications.imagenet_utils import decode_predictions
    from keras.applications.inception_v3 import InceptionV3, preprocess_input
    import dtlpy as dl

    ##############
    # load model #
    ##############
    model = InceptionV3()

    ##########################
    # init platform instance #
    ##########################
    project = dl.projects.get(project_name='ImageNet')
    dataset = project.datasets.get(dataset_name='sample')

    # get pages of images from dataset
    pages = dataset.items.list()
    ####################
    # start annotating #
    ####################
    for page in pages:
        for item in page:
            if item.type == 'dir':
                continue
            img_batch = [item.download(save_locally=False)]
            # load images
            img_batch = [Image.open(buf) for buf in img_batch]
            # get original images shapes before reshaping for model
            orig_img_shape = [img.size[::-1] for img in img_batch]
            # reshape and load images
            batch = np.array([np.array(img.resize((299, 299))) for img in img_batch])
            # preprocess batch
            batch = preprocess_input(batch)
            # inference the model
            predictions = model.predict(batch)
            # get ImageNet labels
            labels = decode_predictions(predictions, top=1)
            # create platform annotations instance
            builder = item.annotations.builder()
            for i_pred, label in enumerate(labels):

                # add the class labels
                ##############################
                # If model is classification #
                ##############################
                builder.add(annotation_definition=dl.Classification(label=label[0][1]))
                #############################
                # If model outputs polygons #
                #############################
                builder.add(annotation_definition=dl.Polyline(geo=pred['polygon_pts'],
                                                              label=labels[i_pred][0][1]))
                #########################
                # If model outputs mask #
                #########################
                builder.add(annotation_definition=dl.Segmentation(geo=pred['mask'],
                                                                  label=labels[i_pred][0][1]))
            # upload a annotations to matching items in platform
            builder.upload()




class BaseCache(object):
    def set(self, key, value):
        """
        set or add a key and value to the cache
        :param key: str or int type of key
        :param value: pickled value
        :return:
        """
        pass

    def get(self, key):
        """
        get the value of the key from the cache
        :param key: str or int type of key
        :return: the value of the key
        """
        pass

    def delete(self, key):
        """
        delete the element from the cache
        :param key: str or int type of key
        :return:
        """
        pass

    def keys(self):
        """
        return all the cache keys
        :return: list of the keys
        """
        pass


import attr
from .. import entities
import logging

logger = logging.getLogger(name='dtlpy')


@attr.s
class TimeSeries(entities.BaseEntity):
    """
    Time Series object
    """
    # platform
    created_at = attr.ib()
    updated_at = attr.ib(repr=False)
    owner = attr.ib()
    name = attr.ib()
    id = attr.ib()
    # entities
    _project = attr.ib(repr=False)

    @property
    def createdAt(self):
        return self.created_at

    @property
    def updatedAt(self):
        return self.updated_at

    @property
    def project(self):
        assert isinstance(self._project, entities.Project)
        return self._project

    @classmethod
    def from_json(cls, _json, project):
        """
        Build a TimeSeries entity object from a json

        :param _json: _json response from host
        :param project: project id
        :return: Time Series object
        """
        return cls(
            created_at=_json.get('createdAt', None),
            updated_at=_json.get('updatedAt', None),
            owner=_json.get('owner', None),
            name=_json.get('name', None),
            id=_json.get('id', None),
            project=project,
        )

    def to_json(self):
        """
        Returns platform _json format of object

        :return: platform json format of object
        :rtype: dict
        """
        _json = attr.asdict(self,
                            filter=attr.filters.exclude(attr.fields(TimeSeries)._project,
                                                        attr.fields(TimeSeries).created_at,
                                                        attr.fields(TimeSeries).updated_at,
                                                        ))
        _json['createdAt'] = self.created_at
        _json['updatedAt'] = self.updated_at
        return _json

    ##########
    # Entity #
    ##########
    def delete(self):
        """
        delete the time series

        :return:
        """
        return self.project.times_series.delete(series=self)

    ##########
    # Series #
    ##########
    def samples(self, filters=None):
        """
        get the time table according to filters

        :param dtlpy.entities.filters.Filters filters: Filters entity or a dictionary containing filters parameters
        :return:
        """
        return self.project.times_series.get_samples(series_id=self.id, filters=filters)

    def add_samples(self, data):
        """
        add data to time series table

        :param data: data
        :return:
        """
        return self.project.times_series.add_samples(series_id=self.id, data=data)

    def delete_samples(self, filters):
        """
        add data to time series table

        :param dtlpy.entities.filters.Filters filters: Filters entity or a dictionary containing filters parameters
        :return:
        """
        return self.project.times_series.delete_samples(series_id=self.id, filters=filters)

    ###########
    # Samples #
    ###########
    def sample(self, sample_id):
        """
        get a sample line by id

        :param sample_id: sample id
        :return:
        """
        return self.project.times_series.get_sample(series_id=self.id,
                                                    sample_id=sample_id)

    def update_sample(self, sample_id, data):
        """
        Update a sample line by id

        :param sample_id: sample id
        :param data: data
        :return:
        """
        return self.project.times_series.update_sample(series_id=self.id,
                                                       sample_id=sample_id,
                                                       data=data)

    def delete_sample(self, sample_id):
        """
        Delete a single sample line from time series

        :param sample_id:sample id
        :return:
        """
        return self.project.times_series.delete_sample(series_id=self.id,
                                                       sample_id=sample_id)



import math
import numpy as np

from . import BaseAnnotationDefinition


class Cube(BaseAnnotationDefinition):
    """
        Cube annotation object
    """
    type = "cube"

    def __init__(self, label, front_tl, front_tr, front_br, front_bl,
                 back_tl, back_tr, back_br, back_bl, angle=None,
                 attributes=None, description=None):
        super().__init__(description=description, attributes=attributes)
        self.front_bl = front_bl
        self.front_br = front_br
        self.front_tr = front_tr
        self.front_tl = front_tl
        self.back_bl = back_bl
        self.back_br = back_br
        self.back_tr = back_tr
        self.back_tl = back_tl

        self._angle = angle
        self.label = label

        self.keys = ["front_tl", "front_tr", "front_br", "front_bl",
                     "back_tl", "back_tr", "back_br", "back_bl"]

    @staticmethod
    def calculate_angle(b, c):
        a = [b[0] + 200, b[1]]
        if b in (a, c):
            return 0

        ang = math.degrees(
            math.atan2(c[1] - b[1], c[0] - b[0]) - math.atan2(a[1] - b[1], a[0] - b[0]))
        return ang + 360 if ang < 0 else ang

    @property
    def angle(self):
        if self._angle is None:
            self._angle = Cube.calculate_angle(self.front_tl, self.front_tr)
        return self._angle

    @property
    def x(self):
        return self.geo[:, 0]

    @property
    def y(self):
        return self.geo[:, 1]

    @property
    def geo(self):
        return np.asarray([self.front_tl,
                           self.front_tr,
                           self.front_br,
                           self.front_bl,
                           self.back_tl,
                           self.back_tr,
                           self.back_br,
                           self.back_bl
                           ])

    def show(self, image, thickness, with_text, height, width, annotation_format, color, alpha=1):
        """
        Show annotation as ndarray
        :param image: empty or image to draw on
        :param thickness:
        :param with_text: not required
        :param height: item height
        :param width: item width
        :param annotation_format: options: list(dl.ViewAnnotationOptions)
        :param color: color
        :param alpha: opacity value [0 1], default 1
        :return: ndarray
        """
        try:
            import cv2
        except (ImportError, ModuleNotFoundError):
            self.logger.error(
                'Import Error! Cant import cv2. Annotations operations will be limited. import manually and fix errors')
            raise

        image = cv2.polylines(image,
                              pts=[np.asarray(
                                  [self.front_bl, self.front_br, self.front_tr, self.front_tl]).round().astype(int)],
                              isClosed=True,
                              color=color,
                              thickness=thickness)
        image = cv2.polylines(image,
                              pts=[np.asarray([self.back_bl, self.back_br, self.back_tr, self.back_tl]).round().astype(
                                  int)],
                              isClosed=True,
                              color=color,
                              thickness=thickness)
        image = cv2.line(image,
                         pt1=tuple(np.asarray(self.front_bl).round().astype(int)),
                         pt2=tuple(np.asarray(self.back_bl).round().astype(int)),
                         color=color,
                         thickness=thickness)
        image = cv2.line(image,
                         pt1=tuple(np.asarray(self.front_br).round().astype(int)),
                         pt2=tuple(np.asarray(self.back_br).round().astype(int)),
                         color=color,
                         thickness=thickness)
        image = cv2.line(image,
                         pt1=tuple(np.asarray(self.front_tl).round().astype(int)),
                         pt2=tuple(np.asarray(self.back_tl).round().astype(int)),
                         color=color,
                         thickness=thickness)
        image = cv2.line(image,
                         pt1=tuple(np.asarray(self.front_tr).round().astype(int)),
                         pt2=tuple(np.asarray(self.back_tr).round().astype(int)),
                         color=color,
                         thickness=thickness)
        return image

    def to_coordinates(self, color):
        coordinates = {self.keys[idx]: {"x": float(x), "y": float(y), "z": 0}
                       for idx, [x, y] in enumerate(self.geo)}
        coordinates['angle'] = self.angle
        return coordinates

    @staticmethod
    def from_coordinates(coordinates):
        geo = list()
        for key, pt in enumerate(coordinates):
            geo.append([pt["x"], pt["y"]])
        return np.asarray(geo)

    @classmethod
    def from_json(cls, _json):
        if "coordinates" in _json:
            key = "coordinates"
        elif "data" in _json:
            key = "data"
        else:
            raise ValueError('can not find "coordinates" or "data" in annotation. id: {}'.format(_json["id"]))

        return cls(
            front_bl=np.asarray([_json[key]["front_bl"]['x'], _json[key]["front_bl"]['y']]),
            front_br=np.asarray([_json[key]["front_br"]['x'], _json[key]["front_br"]['y']]),
            front_tl=np.asarray([_json[key]["front_tl"]['x'], _json[key]["front_tl"]['y']]),
            front_tr=np.asarray([_json[key]["front_tr"]['x'], _json[key]["front_tr"]['y']]),
            back_bl=np.asarray([_json[key]["back_bl"]['x'], _json[key]["back_bl"]['y']]),
            back_br=np.asarray([_json[key]["back_br"]['x'], _json[key]["back_br"]['y']]),
            back_tl=np.asarray([_json[key]["back_tl"]['x'], _json[key]["back_tl"]['y']]),
            back_tr=np.asarray([_json[key]["back_tr"]['x'], _json[key]["back_tr"]['y']]),
            label=_json["label"],
            angle=_json[key]["angle"],
            attributes=_json.get("attributes", None)
        )

    @staticmethod
    def rotate(center, point, angle):
        angle = math.radians(angle)
        cx, cy = center
        px, py = point

        qx = cx + math.cos(angle) * (px - cx) - math.sin(angle) * (py - cy)
        qy = cy + math.sin(angle) * (px - cx) + math.cos(angle) * (py - cy)
        return [qx, qy]

    @classmethod
    def from_boxes_and_angle(cls,
                             front_left, front_top, front_right, front_bottom,
                             back_left, back_top, back_right, back_bottom,
                             label, angle=0, attributes=None):
        """
        Create cuboid by given front and back boxes with angle
        the angle calculate fom the center of each box
        """
        if angle != 0:
            front_center = [front_left + (front_right - front_left) / 2, front_top + (front_bottom - front_top) / 2]
            back_center = [back_left + (back_right - back_left) / 2, back_top + (back_bottom - back_top) / 2]
            front_tl = Cube.rotate(center=front_center, point=[front_left, front_top], angle=angle)
            front_tr = Cube.rotate(center=front_center, point=[front_right, front_top], angle=angle)
            front_br = Cube.rotate(center=front_center, point=[front_right, front_bottom], angle=angle)
            front_bl = Cube.rotate(center=front_center, point=[front_left, front_bottom], angle=angle)
            back_tl = Cube.rotate(center=back_center, point=[back_left, back_top], angle=angle)
            back_tr = Cube.rotate(center=back_center, point=[back_right, back_top], angle=angle)
            back_br = Cube.rotate(center=back_center, point=[back_right, back_bottom], angle=angle)
            back_bl = Cube.rotate(center=back_center, point=[back_left, back_bottom], angle=angle)
        else:
            front_tl = [front_left, front_top]
            front_tr = [front_right, front_top]
            front_br = [front_right, front_bottom]
            front_bl = [front_left, front_bottom]
            back_tl = [back_left, back_top]
            back_tr = [back_right, back_top]
            back_br = [back_right, back_bottom]
            back_bl = [back_left, back_bottom]

        return cls(
            front_tl=front_tl, front_tr=front_tr, front_br=front_br, front_bl=front_bl,
            back_tl=back_tl, back_tr=back_tr, back_br=back_br, back_bl=back_bl,
            label=label,
            angle=angle,
            attributes=attributes
        )



import logging

logger = logging.getLogger(name='dtlpy')


class DirectoryTree:
    """
    Dataset DirectoryTree

    """

    def __init__(self, _json):
        self.dirs = list()
        self.root = SingleDirectory(directory_tree=self, children=_json.get('children', None),
                                    value=_json.get('value', None))
        self.tree = _json

    @property
    def dir_names(self):
        return [directory.path for directory in self.dirs]


class SingleDirectory:
    """
    DirectoryTree single directory

    """

    def __init__(self, value, directory_tree, children=None):
        self.id = value.get('id', None)
        self.name = value.get('name', None)
        self.parent = value.get('dir', None)
        self.path = value.get('filename', None)
        self.metadata = value.get('metadata', dict())

        self.children = list()

        if children is not None:
            for child in children:
                self.children.append(
                    SingleDirectory(directory_tree=directory_tree, children=child.get('children', None),
                                    value=child['value']))

        directory_tree.dirs.append(self)



import threading
import logging
import time
from copy import deepcopy

import numpy as np

from .. import exceptions, entities, repositories, miscellaneous, _api_reference
from ..services.api_client import ApiClient

logger = logging.getLogger(name='dtlpy')
MAX_SLEEP_TIME = 30


class Executions:
    """
    Service Executions Repository

    The Executions class allows the users to manage executions (executions of services) and their properties.
    See our documentation for more information about `executions <https://developers.dataloop.ai/tutorials/faas/execution_control/chapter/>`_.
    """

    def __init__(self,
                 client_api: ApiClient,
                 service: entities.Service = None,
                 project: entities.Project = None):
        self._client_api = client_api
        self._service = service
        self._project = project

    ############
    # entities #
    ############
    @property
    def service(self) -> entities.Service:
        if self._service is None:
            raise exceptions.PlatformException(
                error='2001',
                message='Missing "service". need to set a Service entity or use service.executions repository')
        assert isinstance(self._service, entities.Service)
        return self._service

    @service.setter
    def service(self, service: entities.Service):
        if not isinstance(service, entities.Service):
            raise ValueError('Must input a valid Service entity')
        self._service = service

    @property
    def project(self) -> entities.Project:
        if self._project is None:
            if self._service is not None:
                self._project = self._service._project
        if self._project is None:
            raise exceptions.PlatformException(
                error='2001',
                message='Missing "project". need to set a Project entity or use Project.executions repository')
        assert isinstance(self._project, entities.Project)
        return self._project

    @project.setter
    def project(self, project: entities.Project):
        if not isinstance(project, entities.Project):
            raise ValueError('Must input a valid Project entity')
        self._project = project

    def __get_from_entity(self, name, value):
        project_id = None
        try:
            from_dataset = False
            entity_obj = None
            for input_type in entities.FunctionIO.INPUT_TYPES:
                input_type = input_type.value
                entity = input_type.lower()
                param = '{}_id'.format(entity)
                if isinstance(value, dict):
                    if param in value:
                        repo = getattr(repositories, '{}s'.format(input_type))(client_api=self._client_api)
                        entity_obj = repo.get(**{param: value[param]})
                        if param in ['annotation_id', 'item_id']:
                            from_dataset = True
                            if param == 'item_id':
                                entity_obj = entity_obj.dataset
                            else:
                                entity_obj = repositories.Datasets(client_api=self._client_api).get(
                                    dataset_id=entity_obj.dataset_id)
                        elif param in ['dataset_id']:
                            from_dataset = True
                        break
                elif isinstance(value, str):
                    if entity == name:
                        repo = getattr(repositories, '{}s'.format(input_type))(client_api=self._client_api)
                        entity_obj = repo.get(**{param: value})
                        if name in ['annotation', 'item']:
                            from_dataset = True
                            if param == 'item_id':
                                entity_obj = entity_obj.dataset
                            else:
                                entity_obj = repositories.Datasets(client_api=self._client_api).get(
                                    dataset_id=entity_obj.dataset_id)
                        elif name in ['dataset']:
                            from_dataset = True
                        break

            if entity_obj is not None:
                if isinstance(entity_obj, entities.Project):
                    project_id = entity_obj.id
                elif from_dataset:
                    project_id = entity_obj.projects[0]
                else:
                    project_id = entity_obj.project_id
        except Exception:
            project_id = None

        return project_id

    def __get_project_id(self, project_id=None, payload=None):
        if project_id is None:
            inputs = payload.get('input', dict())
            try:
                for key, val in inputs.items():
                    project_id = self.__get_from_entity(name=key, value=val)
                    if project_id is not None:
                        break
            except Exception:
                project_id = None

            if project_id is None:
                # if still None - get from current repository
                if self._project is not None:
                    project_id = self._project.id
                else:
                    raise exceptions.PlatformException('400', 'Please provide project_id')

        return project_id

    ###########
    # methods #
    ###########
    @_api_reference.add(path='/executions/{serviceId}', method='post')
    def create(self,
               # executions info
               service_id: str = None,
               execution_input: list = None,
               function_name: str = None,
               # inputs info
               resource: entities.PackageInputType = None,
               item_id: str = None,
               dataset_id: str = None,
               annotation_id: str = None,
               project_id: str = None,
               # execution config
               sync: bool = False,
               stream_logs: bool = False,
               return_output: bool = False,
               # misc
               return_curl_only: bool = False,
               timeout: int = None) -> entities.Execution:
        """
        Execute a function on an existing service

        **Prerequisites**: You must be in the role of an *owner* or *developer*. You must have a service.

        :param str service_id: service id to execute on
        :param List[FunctionIO] or dict execution_input: input dictionary or list of FunctionIO entities
        :param str function_name: function name to run
        :param str resource: input type.
        :param str item_id: optional - item id as input to function
        :param str dataset_id: optional - dataset id as input to function
        :param str annotation_id: optional - annotation id as input to function
        :param str project_id: resource's project
        :param bool sync: if true, wait for function to end
        :param bool stream_logs: prints logs of the new execution. only works with sync=True
        :param bool return_output: if True and sync is True - will return the output directly
        :param bool return_curl_only: return the cURL of the creation WITHOUT actually do it
        :param int timeout: int, seconds to wait until TimeoutError is raised. if <=0 - wait until done -
         by default wait take the service timeout
        :return: execution object
        :rtype: dtlpy.entities.execution.Execution

        **Example**:

        .. code-block:: python

            service.executions.create(function_name='function_name', item_id='item_id', project_id='project_id')
        """
        if service_id is None:
            if self._service is None:
                raise exceptions.PlatformException('400', 'Please provide service id')
            service_id = self._service.id

        if resource is None:
            if annotation_id is not None:
                resource = entities.PackageInputType.ANNOTATION
            elif item_id is not None:
                resource = entities.PackageInputType.ITEM
            elif dataset_id is not None:
                resource = entities.PackageInputType.DATASET

        # payload
        payload = dict()
        if execution_input is None:
            if resource is not None:
                inputs = {resource.lower(): {
                    'dataset_id': dataset_id}
                }
                if item_id is not None:
                    inputs[resource.lower()]['item_id'] = item_id
                if annotation_id is not None:
                    inputs[resource.lower()]['annotation_id'] = annotation_id
                payload['input'] = inputs
        else:
            if isinstance(execution_input, dict):
                payload['input'] = execution_input
            else:
                if not isinstance(execution_input, list):
                    execution_input = [execution_input]
                if len(execution_input) > 0 and isinstance(execution_input[0], entities.FunctionIO):
                    payload['input'] = dict()
                    for single_input in execution_input:
                        payload['input'].update(single_input.to_json(resource='execution'))
                else:
                    raise exceptions.PlatformException('400', 'Unknown input type')

        payload['projectId'] = self.__get_project_id(project_id=project_id, payload=payload)

        if function_name is not None:
            payload['functionName'] = function_name
        else:
            payload['functionName'] = entities.package_defaults.DEFAULT_PACKAGE_FUNCTION_NAME

        # request url
        url_path = '/executions/{service_id}'.format(service_id=service_id)

        if return_curl_only:
            curl = self._client_api.export_curl_request(req_type='post',
                                                        path=url_path,
                                                        json_req=payload)
            logger.warning(msg='Execution was NOT created. Exporting cURL only.')
            return curl
        success, response = self._client_api.gen_request(req_type='post',
                                                         path=url_path,
                                                         json_req=payload)
        # exception handling
        if not success:
            raise exceptions.PlatformException(response)

        # return entity
        execution = entities.Execution.from_json(_json=response.json(),
                                                 client_api=self._client_api,
                                                 project=self._project,
                                                 service=self._service)

        if sync and not return_output and not stream_logs:
            execution = self.wait(execution_id=execution.id, timeout=timeout)

        if sync and (stream_logs or return_output):
            thread = None
            if stream_logs:
                thread = threading.Thread(target=self.logs,
                                          kwargs={'execution_id': execution.id,
                                                  'follow': True,
                                                  'until_completed': True})
                thread.setDaemon(True)
                thread.start()
            execution = self.get(execution_id=execution.id,
                                 sync=True)
            # stream logs
            if stream_logs and thread is not None:
                thread.join()
        if sync and return_output:
            return execution.output
        return execution

    @_api_reference.add(path='/executions/{serviceId}', method='post')
    def create_batch(self,
                     service_id: str,
                     filters,
                     function_name: str = None,
                     execution_inputs: list = None,
                     wait=True
                     ):
        """
        Execute a function on an existing service

        **Prerequisites**: You must be in the role of an *owner* or *developer*. You must have a service.

        :param str service_id: service id to execute on
        :param filters: Filters entity for a filtering before execute
        :param str function_name: function name to run
        :param List[FunctionIO] or dict execution_inputs: input dictionary or list of FunctionIO entities, that represent the extra inputs of the function
        :param bool wait: wait until create task finish
        :return: execution object
        :rtype: dtlpy.entities.execution.Execution

        **Example**:

        .. code-block:: python

            command = service.executions.create_batch(
                        execution_inputs=dl.FunctionIO(type=dl.PackageInputType.STRING, value='test', name='string'),
                        filters=dl.Filters(field='dir', values='/test'),
                        function_name='run')
        """
        if service_id is None:
            if self._service is None:
                raise exceptions.PlatformException('400', 'Please provide service id')
            service_id = self._service.id

        if filters is None:
            raise exceptions.PlatformException('400', 'Please provide filter')

        if execution_inputs is None:
            execution_inputs = dict()

        if isinstance(execution_inputs, dict):
            extra_inputs = execution_inputs
        else:
            if not isinstance(execution_inputs, list):
                execution_inputs = [execution_inputs]
            if len(execution_inputs) > 0 and isinstance(execution_inputs[0], entities.FunctionIO):
                extra_inputs = dict()
                for single_input in execution_inputs:
                    extra_inputs.update(single_input.to_json(resource='execution'))
            else:
                raise exceptions.PlatformException('400', 'Unknown input type')

        # payload
        payload = dict()
        payload['batch'] = dict()
        payload['batch']['query'] = filters.prepare()
        payload['batch']['args'] = extra_inputs

        if function_name is not None:
            payload['functionName'] = function_name
        else:
            payload['functionName'] = entities.package_defaults.DEFAULT_PACKAGE_FUNCTION_NAME

        # request url
        url_path = '/executions/{service_id}'.format(service_id=service_id)

        success, response = self._client_api.gen_request(req_type='post',
                                                         path=url_path,
                                                         json_req=payload)
        # exception handling
        if not success:
            raise exceptions.PlatformException(response)

        response_json = response.json()
        command = entities.Command.from_json(_json=response_json,
                                             client_api=self._client_api)
        if wait:
            command = command.wait(timeout=0)
        return command

    @_api_reference.add(path='/executions/rerun', method='post')
    def rerun_batch(self,
                    filters,
                    service_id: str = None,
                    wait=True
                    ):
        """
        rerun a executions on an existing service

        **Prerequisites**: You must be in the role of an *owner* or *developer*. You must have a Filter.

        :param filters: Filters entity for a filtering before rerun
        :param str service_id: service id to rerun on
        :param bool wait: wait until create task finish
        :return: rerun command
        :rtype: dtlpy.entities.command.Command

        **Example**:

        .. code-block:: python

            command = service.executions.rerun_batch(
                        filters=dl.Filters(field='id', values=['executionId'], operator=dl.FiltersOperations.IN, resource=dl.FiltersResource.EXECUTION))
        """
        url_path = '/executions/rerun'

        if filters is None:
            raise exceptions.PlatformException('400', 'Please provide filter')

        if filters.resource != entities.FiltersResource.EXECUTION:
            raise exceptions.PlatformException(
                error='400',
                message='Filters resource must to be FiltersResource.EXECUTION. Got: {!r}'.format(filters.resource))

        if service_id is not None and not filters.has_field('serviceId'):
            filters = deepcopy(filters)
            filters.add(field='serviceId', values=service_id, method=entities.FiltersMethod.AND)

        success, response = self._client_api.gen_request(req_type='post',
                                                         path=url_path,
                                                         json_req={'query': filters.prepare()['filter']})
        # exception handling
        if not success:
            raise exceptions.PlatformException(response)

        response_json = response.json()
        command = entities.Command.from_json(_json=response_json,
                                             client_api=self._client_api)
        if wait:
            command = command.wait(timeout=0)
        return command

    def _list(self, filters: entities.Filters):
        """
        List service executions

        :param dtlpy.entities.filters.Filters filters: dl.Filters entity to filters items
        :return:
        """
        url = '/query/faas'

        # request
        success, response = self._client_api.gen_request(req_type='POST',
                                                         path=url,
                                                         json_req=filters.prepare())
        if not success:
            raise exceptions.PlatformException(response)

        return response.json()

    @_api_reference.add(path='/query/faas', method='post')
    def list(self, filters: entities.Filters = None) -> entities.PagedEntities:
        """
        List service executions

        **Prerequisites**: You must be in the role of an *owner* or *developer*. You must have a service.

        :param dtlpy.entities.filters.Filters filters: dl.Filters entity to filters items
        :return: Paged entity
        :rtype: dtlpy.entities.paged_entities.PagedEntities

        **Example**:

        .. code-block:: python

            service.executions.list()
        """
        # default filtersf
        if filters is None:
            filters = entities.Filters(resource=entities.FiltersResource.EXECUTION)
        # assert type filters
        elif not isinstance(filters, entities.Filters):
            raise exceptions.PlatformException(
                error='400',
                message='Unknown filters type: {!r}'.format(type(filters)))
        if filters.resource != entities.FiltersResource.EXECUTION:
            raise exceptions.PlatformException(
                error='400',
                message='Filters resource must to be FiltersResource.EXECUTION. Got: {!r}'.format(filters.resource))
        if self._project is not None:
            filters.add(field='projectId', values=self._project.id)
        if self._service is not None:
            filters.add(field='serviceId', values=self._service.id)

        paged = entities.PagedEntities(items_repository=self,
                                       filters=filters,
                                       page_offset=filters.page,
                                       page_size=filters.page_size,
                                       client_api=self._client_api)
        paged.get_page()
        return paged

    def _build_entities_from_response(self, response_items) -> miscellaneous.List[entities.Execution]:
        pool = self._client_api.thread_pools(pool_name='entity.create')
        jobs = [None for _ in range(len(response_items))]
        # return execution list
        for i_item, item in enumerate(response_items):
            jobs[i_item] = pool.submit(entities.Execution._protected_from_json,
                                       **{'client_api': self._client_api,
                                          '_json': item,
                                          'project': self._project,
                                          'service': self._service})

        # get results
        results = [j.result() for j in jobs]
        # log errors
        _ = [logger.warning(r[1]) for r in results if r[0] is False]
        # return good jobs
        return miscellaneous.List([r[1] for r in results if r[0] is True])

    @_api_reference.add(path='/executions/{id}', method='get')
    def get(self,
            execution_id: str = None,
            sync: bool = False
            ) -> entities.Execution:
        """
        Get Service execution object

        **Prerequisites**: You must be in the role of an *owner* or *developer*. You must have a service.

        :param str execution_id: execution id
        :param bool sync: if true, wait for the execution to finish
        :return: Service execution object
        :rtype: dtlpy.entities.execution.Execution

        **Example**:

        .. code-block:: python

            service.executions.get(execution_id='execution_id')
        """
        url_path = "/executions/{}".format(execution_id)
        if sync:
            return self.wait(execution_id=execution_id)

        success, response = self._client_api.gen_request(req_type="get",
                                                         path=url_path)

        # exception handling
        if not success:
            raise exceptions.PlatformException(response)

        # return entity
        return entities.Execution.from_json(client_api=self._client_api,
                                            _json=response.json(),
                                            project=self._project,
                                            service=self._service)

    def logs(self,
             execution_id: str,
             follow: bool = True,
             until_completed: bool = True):
        """
        executions logs

        **Prerequisites**: You must be in the role of an *owner* or *developer*. You must have a service.

        :param str execution_id: execution id
        :param bool follow: if true, keep stream future logs
        :param bool until_completed: if true, wait until completed
        :return: executions logs

        **Example**:

        .. code-block:: python

            service.executions.logs(execution_id='execution_id')
        """
        return self.service.log(execution_id=execution_id,
                                follow=follow,
                                until_completed=until_completed,
                                view=True)

    def increment(self, execution: entities.Execution):
        """
        Increment the number of attempts that an execution is allowed to attempt to run a service that is not responding.

        **Prerequisites**: You must be in the role of an *owner* or *developer*. You must have a service.

        :param dtlpy.entities.execution.Execution execution:
        :return: int
        :rtype: int

        **Example**:

        .. code-block:: python

            service.executions.increment(execution='execution_entity')
        """
        # request
        success, response = self._client_api.gen_request(
            req_type='post',
            path='/executions/{}/attempts'.format(execution.id)
        )

        # exception handling
        if not success:
            raise exceptions.PlatformException(response)

        # exception handling
        if not success:
            raise exceptions.PlatformException(response)
        else:
            return response.json()

    @_api_reference.add(path='/executions/{executionId}/rerun', method='post')
    def rerun(self, execution: entities.Execution, sync: bool = False):
        """
        Rerun execution

        **Prerequisites**: You must be in the role of an *owner* or *developer*. You must have a service.

        :param dtlpy.entities.execution.Execution execution:
        :param bool sync: wait for the execution to finish
        :return: Execution object
        :rtype: dtlpy.entities.execution.Execution

        **Example**:

        .. code-block:: python

            service.executions.rerun(execution='execution_entity')
        """

        url_path = "/executions/{}/rerun".format(execution.id)
        # request
        success, response = self._client_api.gen_request(req_type='post',
                                                         path=url_path)

        # exception handling
        if not success:
            raise exceptions.PlatformException(response)
        else:
            execution = entities.Execution.from_json(
                client_api=self._client_api,
                _json=response.json(),
                project=self._project,
                service=self._service
            )
            if sync:
                execution = self.wait(execution_id=execution.id)
        return execution

    def wait(self,
             execution_id: str = None,
             execution: entities.Execution = None,
             timeout: int = None,
             backoff_factor=1):
        """
        Get Service execution object.

        **Prerequisites**: You must be in the role of an *owner* or *developer*. You must have a service.

        :param str execution_id: execution id
        :param str execution: dl.Execution, optional. must input one of execution or execution_id
        :param int timeout: seconds to wait until TimeoutError is raised. if <=0 - wait until done - by default wait take the service timeout
        :param float backoff_factor: A backoff factor to apply between attempts after the second try
        :return: Service execution object
        :rtype: dtlpy.entities.execution.Execution

        **Example**:

        .. code-block:: python

            service.executions.wait(execution_id='execution_id')
        """
        if execution is None:
            if execution_id is None:
                raise ValueError('Must input at least one: [execution, execution_id]')
            else:
                execution = self.get(execution_id=execution_id)
        elapsed = 0
        start = int(time.time())
        if timeout is None or timeout <= 0:
            timeout = np.inf

        num_tries = 1
        while elapsed < timeout:
            execution = self.get(execution_id=execution.id)
            if not execution.in_progress():
                break
            elapsed = time.time() - start
            if elapsed >= timeout:
                raise TimeoutError(
                    f"execution wait() got timeout. id: {execution.id!r}, status: {execution.latest_status}")
            sleep_time = np.min([timeout - elapsed, backoff_factor * (2 ** num_tries), MAX_SLEEP_TIME])
            num_tries += 1
            logger.debug(f"Execution {execution.id} is running for {elapsed:.2f}[s]. Sleeping for {sleep_time:.2f}[s]")
            time.sleep(sleep_time)

        return execution

    @_api_reference.add(path='/executions/{id}/terminate', method='post')
    def terminate(self, execution: entities.Execution):
        """
        Terminate Execution

        **Prerequisites**: You must be in the role of an *owner* or *developer*. You must have a service.

        :param dtlpy.entities.execution.Execution execution:
        :return: execution object
        :rtype: dtlpy.entities.execution.Execution

        **Example**:

        .. code-block:: python

            service.executions.terminate(execution='execution_entity')
        """
        # request
        success, response = self._client_api.gen_request(req_type='post',
                                                         path='/executions/{}/terminate'.format(execution.id))

        # exception handling
        if not success:
            raise exceptions.PlatformException(response)
        else:
            return entities.Execution.from_json(_json=response.json(),
                                                service=self._service,
                                                project=self._project,
                                                client_api=self._client_api)

    @_api_reference.add(path='/executions/{id}', method='patch')
    def update(self, execution: entities.Execution) -> entities.Execution:
        """
        Update execution changes to platform

        **Prerequisites**: You must be in the role of an *owner* or *developer*. You must have a service.

        :param dtlpy.entities.execution.Execution execution: execution entity
        :return: Service execution object
        :rtype: dtlpy.entities.execution.Execution

        **Example**:

        .. code-block:: python

            service.executions.update(execution='execution_entity')
        """
        # payload
        payload = execution.to_json()

        # request
        success, response = self._client_api.gen_request(req_type='patch',
                                                         path='/executions/{}'.format(execution.id),
                                                         json_req=payload)

        # exception handling
        if not success:
            raise exceptions.PlatformException(response)

        # return entity
        if self._project is not None:
            project = self._project
        else:
            project = execution._project

        # return
        if self._service is not None:
            service = self._service
        else:
            service = execution._service

        return entities.Execution.from_json(_json=response.json(),
                                            service=service,
                                            project=self._project,
                                            client_api=self._client_api)

    def progress_update(
            self,
            execution_id: str,
            status: entities.ExecutionStatus = None,
            percent_complete: int = None,
            message: str = None,
            output: str = None,
            service_version: str = None
    ):
        """
        Update Execution Progress.

        **Prerequisites**: You must be in the role of an *owner* or *developer*. You must have a service.

        :param str execution_id: execution id
        :param str status: ExecutionStatus
        :param int percent_complete: percent work done
        :param str message: message
        :param str output: the output of the execution
        :param str service_version: service version
        :return: Service execution object
        :rtype: dtlpy.entities.execution.Execution

        **Example**:

        .. code-block:: python

            service.executions.progress_update(execution_id='execution_id', status='complete', percent_complete=100)
        """
        # create payload
        payload = dict()

        if status is not None:
            payload['status'] = status
        else:
            if percent_complete is not None and isinstance(percent_complete, int):
                if percent_complete < 100:
                    payload['status'] = 'inProgress'
                else:
                    payload['status'] = 'completed'
            elif output is not None:
                payload['status'] = 'completed'
            else:
                payload['status'] = 'inProgress'

        if percent_complete is not None:
            payload['percentComplete'] = percent_complete

        if message is not None:
            payload['message'] = message

        if output is not None:
            payload['output'] = output

        if service_version is not None:
            payload['serviceVersion'] = service_version

        # request
        success, response = self._client_api.gen_request(
            req_type="post",
            path="/executions/{}/progress".format(execution_id),
            json_req=payload
        )

        # exception handling
        if success:
            return entities.Execution.from_json(_json=response.json(),
                                                client_api=self._client_api,
                                                project=self._project,
                                                service=self._service)
        else:
            raise exceptions.PlatformException(response)



import json
import logging
import os
from typing import List, Optional
from pathlib import Path

from .. import exceptions, entities, services, miscellaneous, assets
from ..services.api_client import ApiClient

logger = logging.getLogger(name='dtlpy')


class Dpks:
    def __init__(self, client_api: ApiClient, project: entities.Project = None):
        self._client_api = client_api
        self._project = project

    @property
    def project(self) -> Optional[entities.Project]:
        return self._project

    @project.setter
    def project(self, project: entities.Project):
        if not isinstance(project, entities.Project):
            raise ValueError('Must input a valid Project entity')
        self._project = project

    def init(self, directory: str = None, name: str = None, description: str = None,
             attributes: dict = None, icon: str = None, scope: str = None):
        """
        Initialize a dpk project with the specified projects.

        :param str directory: the directory where to initialize the project
        :param str name: the name of the dpk.
        :param str description: the description of the dpk.
        :param str attributes: the attributes of the dpk.
        :param str icon: the icon of the dpk.
        :param str scope: the scope of the dpk.

        ** Example **
        .. code-block:: python
            dl.dpks.init(name='Hello World', description='A description of the dpk', attributes={
                "Provider": "Dataloop",
                "License": "",
                "Category": "Model",
                "Computer Vision": "Object Detection",
                "Media Type": "Image"
              },
                        icon='path_to_icon', scope='organization')
        """
        if directory is None:
            directory = os.getcwd()
        dpk = entities.Dpk.from_json(_json={'name': miscellaneous.JsonUtils.get_if_absent(name),
                                            'description': miscellaneous.JsonUtils.get_if_absent(description),
                                            'attributes': miscellaneous.JsonUtils.get_if_absent(attributes),
                                            'icon': miscellaneous.JsonUtils.get_if_absent(icon),
                                            'scope': miscellaneous.JsonUtils.get_if_absent(scope, 'organization'),
                                            'components': dict()
                                            },
                                     client_api=self._client_api)
        dataloop_filepath = os.path.join(directory, assets.paths.APP_JSON_FILENAME)
        with open(dataloop_filepath, 'w') as json_file:
            json_file.write(json.dumps(dpk.to_json(), indent=4))
        os.makedirs(os.path.join(directory, 'functions'), exist_ok=True)
        os.makedirs(os.path.join(directory, 'panels'), exist_ok=True)

    # noinspection PyMethodMayBeStatic
    def pack(self, directory: str = None, name: str = None, subpaths_to_append: List[str] = None) -> str:
        """
        :param str directory: optional - the project to pack, if not specified use the current project,
        :param str name: optional - the name of the dpk file.
        :param List[str] subpaths_to_append: optional - the files/directories to add to the dpk file.
                                            (along with functions, panels and dataloop.json)
        :return the path of the dpk file

        **Example**

        .. code-block:: python
            filepath = dl.apps.pack(directory='/my-current-project', name='project-name')
        """
        # create/get .dataloop dir
        cwd = os.getcwd()
        dl_dir = os.path.join(cwd, '.dataloop')
        if not os.path.isdir(dl_dir):
            os.mkdir(dl_dir)
        if directory is None:
            directory = cwd

        # get dpk name
        if name is None:
            name = os.path.basename(directory)

        with open(os.path.join(directory, 'dataloop.json'), 'r') as file:
            app_json = json.load(file)
        version = app_json.get('version', None)
        if version is None:
            logger.warning('No Version specified, setting to 1.0.0')
            version = '1.0.0'
        # create/get dist folder
        dpk_filename = os.path.join(dl_dir, '{}_{}.dpk'.format(name, version))

        if not os.path.isdir(directory):
            raise ValueError('Not a directory: {}'.format(directory))
        if subpaths_to_append is None:
            subpaths_to_append = []

        try:
            directory = os.path.abspath(directory)
            # create zipfile
            miscellaneous.Zipping.zip_directory(zip_filename=dpk_filename,
                                                directory=directory,
                                                ignore_directories=['artifacts'])
            return dpk_filename
        except Exception:
            logger.error('Error when packing:')
            raise

    def _build_entities_from_response(self, response_items) -> miscellaneous.List[entities.Dpk]:
        pool = self._client_api.thread_pools(pool_name='entity.create')
        jobs = [None for _ in range(len(response_items))]
        # return triggers list
        for i_item, item in enumerate(response_items):
            jobs[i_item] = pool.submit(entities.Dpk._protected_from_json,
                                       **{'client_api': self._client_api,
                                          '_json': item,
                                          'project': self._project})
        # get all results
        results = [j.result() for j in jobs]
        # log errors
        _ = [logger.warning(r[1]) for r in results if r[0] is False]
        # return good jobs
        items = miscellaneous.List([r[1] for r in results if r[0] is True])
        return items

    def pull(self, dpk: entities.Dpk = None, dpk_id: str = None, dpk_name: str = None, local_path=None) -> str:
        """
        Pulls the app from the platform as dpk file.

        Note: you must pass either dpk_name or dpk_id to the function.
        :param str dpk: DPK entity.
        :param str dpk_id: the name of the dpk.
        :param str dpk_name: the id of the dpk.
        :param local_path: the path where you want to install the dpk file.
        :return local path where the package pull

        **Example**
        ..code-block:: python
            path = dl.dpks.pull(dpk_name='my-app')
        """
        if dpk is None:
            dpk = self.get(dpk_id=dpk_id, dpk_name=dpk_name)
        if local_path is None:
            local_path = os.path.join(
                services.service_defaults.DATALOOP_PATH,
                "dpks",
                dpk.name,
                str(dpk.version))

        dpk.codebases.unpack(codebase=dpk.codebase, local_path=local_path)
        return local_path

    def __get_by_name(self, dpk_name: str, dpk_version: str = None):
        filters = entities.Filters(field='name',
                                   values=dpk_name,
                                   resource=entities.FiltersResource.DPK,
                                   use_defaults=False)
        dpks = self.list(filters=filters)
        # only latest version returns so should be only one
        if dpks.items_count == 0:
            raise exceptions.PlatformException(error='404',
                                               message='Dpk not found. Name: {}'.format(dpk_name))
        elif dpks.items_count > 1:
            raise exceptions.PlatformException(
                error='400',
                message='More than one dpk found by the name of: {}'.format(dpk_name))
        dpk: entities.Dpk = dpks.items[0]

        ########################
        # get specific version #
        ########################
        if dpk_version is not None and dpk.version != dpk_version:
            filters = entities.Filters(field='version',
                                       values=dpk_version,
                                       resource=entities.FiltersResource.DPK,
                                       use_defaults=False)
            dpk_v = dpk._get_revision_pages(filters=filters)
            if dpk_v.items_count == 0:
                raise exceptions.PlatformException(
                    error='404',
                    message=f'Dpk version not found. Name: {dpk_name}, version: {dpk_version}')
            elif dpk_v.items_count > 1:
                # should never be here - more than one with same name and version
                raise exceptions.PlatformException(
                    error='400',
                    message=f'More than one dpk found with name: {dpk_name}, version: {dpk_version}')
            dpk = dpk_v.items[0]
        return dpk

    def publish(
            self,
            dpk: entities.Dpk = None,
            ignore_max_file_size: bool = False,
            manifest_filepath='dataloop.json',
            local_path: str = None
    ) -> entities.Dpk:
        """
        Upload a dpk entity to the dataloop platform.

        :param entities.Dpk dpk: Optional. The DPK entity to publish. If None, a new DPK is created
                             from the manifest file.
        :param bool ignore_max_file_size: Optional. If True, the maximum file size check is ignored
                                        during the packaging of the codebase.
        :param str manifest_filepath: Optional. Path to the manifest file. Can be absolute or relative.
                                    Defaults to 'dataloop.json'
        :param str local_path: Optional. The path where the dpk files are located.

        :return the published dpk
        :rtype dl.entities.Dpk

        **Example**

        .. code-block:: python
            published_dpk = dl.dpks.publish()
        """
        manifest_path = Path(manifest_filepath).resolve()

        if dpk is None:
            if not manifest_path.exists():
                raise FileNotFoundError(f'{manifest_filepath} file must exist in order to publish a dpk')
            with open(manifest_filepath, 'r') as f:
                json_file = json.load(f)
            dpk = entities.Dpk.from_json(_json=json_file,
                                         client_api=self._client_api,
                                         project=self.project)

        if not dpk.context:
            dpk.context = {}
        if 'project' not in dpk.context:
            if not self.project:
                raise exceptions.PlatformException('400', 'project id must be provided in the context')
            dpk.context['project'] = self.project.id
        if 'org' not in dpk.context and dpk.scope == 'organization':
            if not self.project:
                raise exceptions.PlatformException('400', 'org id must be provided in the context')
            dpk.context['org'] = self.project.org['id']

        if self.project and self.project.id != dpk.context['project']:
            logger.warning("the project id that provide different from the dpk project id")

        if local_path is None:
            if manifest_filepath == 'dataloop.json':
                local_path = os.getcwd()
            else:
                local_path = os.path.dirname(manifest_filepath)
        if dpk.codebase is None:
            dpk.codebase = self.project.codebases.pack(directory=local_path,
                                                       name=dpk.name,
                                                       extension='dpk',
                                                       ignore_directories=['artifacts'],
                                                       ignore_max_file_size=ignore_max_file_size)

        success_pack, response_pack = self._client_api.gen_request(req_type='post',
                                                                   json_req=dpk.to_json(),
                                                                   path='/app-registry')
        if not success_pack:
            raise exceptions.PlatformException(error=response_pack)

        return entities.Dpk.from_json(response_pack.json(), self._client_api, dpk.project)

    def update(self, dpk: entities.Dpk) -> entities.Dpk:
        """
        Update the dpk entity in the platform.

        Note: the update will update only attributes, displayName, Icon, description and pipeline template preview

        :param entities.Dpk dpk: the dpk entity to update.
        :return the updated dpk entity.
        :rtype entities.Dpk
        """
        success, response = self._client_api.gen_request(req_type='patch',
                                                         path='/app-registry/{}'.format(dpk.id),
                                                         json_req=dpk.to_json())
        if not success:
            raise exceptions.PlatformException(response)
        res = response.json()
        logger.info(res.get('message'))
        return entities.Dpk.from_json(res.get('dpk'), self._client_api, self._project)

    def delete(self, dpk_id: str) -> bool:
        """
        Delete the dpk from the app store.

        Note: after removing the dpk, you cant get it again, it's advised to pull it first.

        :param str dpk_id: the id of the dpk.
        :return whether the operation ran successfully
        :rtype bool
        """
        success, response = self._client_api.gen_request(req_type='delete', path=f'/app-registry/{dpk_id}')
        if success:
            logger.info(f'Deleted dpk: {dpk_id} successfully')
        else:
            raise exceptions.PlatformException(response)
        return success

    def list(self, filters: entities.Filters = None) -> entities.PagedEntities:
        """
        List the available dpks.

        :param entities.Filters filters: the filters to apply on the list
        :return a paged entity representing the list of dpks.

        ** Example **
        .. code-block:: python
            dpks = dl.dpks.list()
        """
        if filters is None:
            filters = entities.Filters(resource=entities.FiltersResource.DPK)
        elif not isinstance(filters, entities.Filters):
            raise ValueError('Unknown filters type: {!r}'.format(type(filters)))
        elif filters.resource != entities.FiltersResource.DPK:
            raise ValueError('Filters resource must to be FiltersResource.DPK. Got: {!r}'.format(filters.resource))

        if self._project is not None:
            filters.add(field='context.project', values=self._project.id)

        paged = entities.PagedEntities(items_repository=self,
                                       filters=filters,
                                       page_offset=filters.page,
                                       page_size=filters.page_size,
                                       client_api=self._client_api)
        paged.get_page()
        return paged

    def _list(self, filters: entities.Filters):
        url = '/app-registry/query'

        # request
        success, response = self._client_api.gen_request(req_type='post',
                                                         path=url,
                                                         json_req=filters.prepare())
        if not success:
            raise exceptions.PlatformException(response)
        return response.json()

    def get_revisions(self, dpk_id: str, version: str):
        """
        Get the revision of a specific dpk.

        :param str dpk_id: the id of the dpk.
        :param str version: the version of the dpk.
        :return the entity of the dpk
        :rtype entities.Dpk

        ** Example **
        ..coed-block:: python
            dpk = dl.dpks.get_revisions(dpk_id='id', version='1.0.0')
        """
        if dpk_id is None or version is None:
            raise ValueError('You must provide both dpk_id and version')
        url = '/app-registry/{}/revisions/{}'.format(dpk_id, version)

        # request
        success, response = self._client_api.gen_request(req_type='get',
                                                         path=url)
        if not success:
            raise exceptions.PlatformException(response)

        dpk = entities.Dpk.from_json(_json=response.json(),
                                     client_api=self._client_api,
                                     project=self._project,
                                     is_fetched=False)
        return dpk

    def get(self, dpk_name: str = None, dpk_version: str = None, dpk_id: str = None) -> entities.Dpk:
        """
        Get a specific dpk from the platform.

        Note: you must pass either dpk_id or dpk_name.

        :param str dpk_id: the id of the dpk to get.
        :param str dpk_name: the name of the dpk to get.
        :param str dpk_version: options - to get a specific dpk version
        :return the entity of the dpk
        :rtype entities.Dpk

        ** Example **
        ..coed-block:: python
            dpk = dl.dpks.get(dpk_name='name')
        """
        if dpk_id is None and dpk_name is None:
            raise ValueError('You must provide an identifier, either dpk_id or dpk_name')
        if dpk_id is not None:
            url = '/app-registry/{}'.format(dpk_id)

            # request
            success, response = self._client_api.gen_request(req_type='get',
                                                             path=url)
            if not success:
                raise exceptions.PlatformException(response)

            dpk = entities.Dpk.from_json(_json=response.json(),
                                         client_api=self._client_api,
                                         project=self._project,
                                         is_fetched=False)
        else:
            dpk = self.__get_by_name(dpk_name=dpk_name, dpk_version=dpk_version)

        return dpk

    def get_previews(self, dpk: entities.Dpk):
        """
        Get the preview of a specific dpk.

        :param entities.Dpk dpk: the dpk entity to get the preview for.
        :return the preview of the dpk's templates
        :rtype dict

        ** Example **
        ..coed-block:: python
            res = dl.dpks.get_previews(dpk=dpk)
        """
        url = '/app-registry/{}/previews'.format(dpk.id)

        # request
        success, response = self._client_api.gen_request(
            req_type='get',
            path=url
        )

        if not success:
            raise exceptions.PlatformException(response)

        return response.json()


import traceback
import logging
import attr

from .. import entities, exceptions, repositories

logger = logging.getLogger(name='dtlpy')


@attr.s
class Bot(entities.User):
    """
    Bot entity
    """
    _bots = attr.ib(repr=False, default=None)
    password = attr.ib(default=None)

    @staticmethod
    def _protected_from_json(_json, project, client_api, bots=None):
        """
        Same as from_json but with try-except to catch if error
        :param _json: platform json
        :param project: project entity
        :param client_api: ApiClient entity
        :param bots:
        :return:
        """
        try:
            bot = Bot.from_json(_json=_json,
                                project=project,
                                bots=bots,
                                client_api=client_api)
            status = True
        except Exception:
            bot = traceback.format_exc()
            status = False
        return status, bot

    @classmethod
    def from_json(cls, _json, project, client_api, bots=None):
        """
        Build a Bot entity object from a json

        :param _json: _json response from host
        :param project: project entity
        :param client_api: ApiClient entity
        :param bots: Bots repository
        :return: User object
        """
        return cls(
            created_at=_json.get('createdAt', None),
            name=_json.get('firstName', None),
            updated_at=_json.get('updatedAt', None),
            last_name=_json.get('lastName', None),
            username=_json.get('username', None),
            avatar=_json.get('avatar', None),
            email=_json.get('email', None),
            type=_json.get('type', None),
            role=_json.get('role', None),
            org=_json.get('org', None),
            id=_json.get('id', None),
            project=project,
            client_api=client_api,
            bots=bots,
            password=_json.get('password', None),
        )

    @property
    def bots(self):
        if self._bots is None:
            self._bots = repositories.Bots(client_api=self._client_api, project=self._project)
        assert isinstance(self._bots, repositories.Bots)
        return self._bots

    @property
    def project(self):
        if self._project is None:
            raise exceptions.PlatformException(error='2001',
                                               message='Missing entity "project".')
        assert isinstance(self._project, entities.Project)
        return self._project

    def to_json(self):
        """
        Returns platform _json format of object

        :return: platform json format of object
        :rtype: dict
        """
        _json = attr.asdict(self,
                            filter=attr.filters.exclude(attr.fields(Bot)._project,
                                                        attr.fields(Bot).name,
                                                        attr.fields(Bot)._client_api,
                                                        attr.fields(Bot)._bots,
                                                        attr.fields(Bot)._users,
                                                        attr.fields(Bot).last_name,
                                                        attr.fields(Bot).created_at,
                                                        attr.fields(Bot).updated_at,
                                                        ))
        _json['firstName'] = self.name
        _json['lastName'] = self.last_name
        _json['createdAt'] = self.created_at
        _json['updatedAt'] = self.updated_at
        return _json

    def delete(self):
        """
        Delete the bot

        :return: True
        :rtype: bool
        """
        return self.bots.delete(bot_id=self.id)



import threading
import time
import traceback
import logging
import queue
import os

logger = logging.getLogger(name='dtlpy')


class Events(threading.Thread):
    def __init__(self, client_api, *args, **kwargs):
        super(Events, self).__init__(*args, **kwargs)
        self.client_api = client_api
        self.q = queue.Queue()
        self.mapping_events_dict = {
            'project': {'method': ['create', 'delete'], 'route': '/projects'},
            'task': {'method': ['create'], 'route': '/annotationtasks'},
        }

    def track(self, event):
        try:
            return_type, resp = self.client_api.gen_request(req_type='POST',
                                                            path='/analytics/metric/pendo',
                                                            json_req=event,
                                                            log_error=False)
            if not resp.ok:
                logger.debug('failed send event to analytics: {}'.format(resp.text))
        except Exception:
            logger.debug('failed send track event: {}'.format(traceback.format_exc()))

    def run(self):
        while True:
            try:
                event = self.q.get()
                self.track([event])
                self.q.task_done()
            except Exception:
                logger.exception('failed in loop')

    def _valid_events(self, path):
        for route in self.mapping_events_dict.values():
            if path.startswith(route['route']) and 'sdk' not in path:
                return True
        return False

    def _add_info(self, event_payload, function, resp):
        if function in ['create']:
            event_source = event_payload.get('event', None)
            resp_json = resp.json()
            if event_source == 'dtlpy:project':
                event_payload['properties'].update({'project_id': resp_json['id'],
                                                    'project_name': resp_json['name']})
            if event_source == 'dtlpy:task' and function in ['create']:
                if 'createTaskPayload' in resp_json.get('spec', {}):
                    task_payload = resp_json.get('spec', {}).get('createTaskPayload', {})
                else:
                    task_payload = resp_json
                metadata = task_payload.get('metadata', {}).get('system', {})
                task_type = task_payload.get('spec', {}).get('type', {})
                allocation_method = 'Distribution'
                if 'batchSize' in metadata and 'maxBatchWorkload' in metadata and 'allowedAssignees' in metadata:
                    allocation_method = 'Pulling'
                event_payload['properties'].update({'task_type': task_type,
                                                    'allocation_method': allocation_method})

    def put(self, event, resp=None, path=None):
        send_event = True
        if path is not None and not self._valid_events(path=path):
            send_event = False

        if resp is not None and send_event:
            event_source = os.path.normpath(event.filename).split('\\')[-1][:-4]
            event_payload = {'event': 'dtlpy:' + event_source,
                             'properties': {'sdk_event': event.function + '_' + event_source}}
            if event_source in self.mapping_events_dict and \
                    event.function in self.mapping_events_dict[event_source]['method']:
                self._add_info(event_payload=event_payload, function=event.function, resp=resp)
            else:
                send_event = False
        else:
            event_payload = event
        if send_event:
            self.q.put(event_payload)



from collections import namedtuple
import logging
import traceback
from enum import Enum
from typing import List
import attr
from .node import PipelineNode, PipelineConnection, TaskNode, CodeNode, FunctionNode, DatasetNode
from .. import repositories, entities
from ..services.api_client import ApiClient
from .package_function import PackageInputType
import copy

logger = logging.getLogger(name='dtlpy')


class PipelineResumeOption(str, Enum):
    TERMINATE_EXISTING_CYCLES = 'terminateExistingCycles',
    RESUME_EXISTING_CYCLES = 'resumeExistingCycles'


class CompositionStatus(str, Enum):
    CREATED = "Created",
    INITIALIZING = "Initializing",
    INSTALLED = "Installed",
    ACTIVATED = "Activated",
    DEACTIVATED = "Deactivated",
    UNINSTALLED = "Uninstalled",
    TERMINATING = "Terminating",
    TERMINATED = "Terminated",
    UPDATING = "Updating",
    FAILURE = "Failure"


class PipelineSettings:

    def __init__(
            self,
            default_resume_option: PipelineResumeOption = None,
            keep_triggers_active: bool = None,
            active_trigger_ask_again: bool = None,
            last_update: dict = None
    ):
        self.default_resume_option = default_resume_option
        self.keep_triggers_active = keep_triggers_active
        self.active_trigger_ask_again = active_trigger_ask_again
        self.last_update = last_update

    @classmethod
    def from_json(cls, _json: dict = None):
        if _json is None:
            _json = dict()
        return cls(
            default_resume_option=_json.get('defaultResumeOption', None),
            keep_triggers_active=_json.get('keepTriggersActive', None),
            active_trigger_ask_again=_json.get('activeTriggerAskAgain', None),
            last_update=_json.get('lastUpdate', None)
        )

    def to_json(self):
        _json = dict()

        if self.default_resume_option is not None:
            _json['defaultResumeOption'] = self.default_resume_option

        if self.default_resume_option is not None:
            _json['keepTriggersActive'] = self.default_resume_option

        if self.default_resume_option is not None:
            _json['activeTriggerAskAgain'] = self.default_resume_option

        if self.default_resume_option is not None:
            _json['lastUpdate'] = self.default_resume_option

        return _json


class Variable(entities.DlEntity):
    """
    Pipeline Variables
    """
    id: str = entities.DlProperty(location=['id'], _type=str)
    created_at: str = entities.DlProperty(location=['createdAt'], _type=str)
    updated_at: str = entities.DlProperty(location=['updatedAt'], _type=str)
    reference: str = entities.DlProperty(location=['reference'], _type=str)
    creator: str = entities.DlProperty(location=['creator'], _type=str)
    variable_type: PackageInputType = entities.DlProperty(location=['type'], _type=PackageInputType)
    name: str = entities.DlProperty(location=['name'], _type=str)
    value = entities.DlProperty(location=['value'])

    @classmethod
    def from_json(cls, _json):
        """
        Turn platform representation of variable into a pipeline variable entity

        :param dict _json: platform representation of pipeline variable
        :return: pipeline variable entity
        :rtype: dtlpy.entities.pipeline.PipelineVariables
        """

        inst = cls(_dict=_json)
        return inst

    def to_json(self):
        """
        :return: variable of pipeline
        :rtype: dict
        """
        _json = self._dict.copy()
        return _json


class PipelineAverages:
    def __init__(
            self,
            avg_time_per_execution: float,
            avg_execution_per_day: float
    ):
        self.avg_time_per_execution = avg_time_per_execution
        self.avg_execution_per_day = avg_execution_per_day

    @classmethod
    def from_json(cls, _json: dict = None):
        if _json is None:
            _json = dict()
        return cls(
            avg_time_per_execution=_json.get('avgTimePerExecution', 'NA'),
            avg_execution_per_day=_json.get('avgExecutionsPerDay', 'NA')
        )


class NodeAverages:
    def __init__(
            self,
            node_id: str,
            averages: PipelineAverages
    ):
        self.node_id = node_id
        self.averages = averages

    @classmethod
    def from_json(cls, _json: dict):
        return cls(
            node_id=_json.get('nodeId', None),
            averages=PipelineAverages.from_json(_json.get('executionStatistics'))
        )


class PipelineCounter:
    def __init__(
            self,
            status: str,
            count: int
    ):
        self.status = status
        self.count = count


class NodeCounters:
    def __init__(
            self,
            node_id: str,
            counters: List[PipelineCounter]
    ):
        self.node_id = node_id
        self.counters = counters

    @classmethod
    def from_json(cls, _json: dict):
        return cls(
            node_id=_json.get('nodeId', None),
            counters=[PipelineCounter(**c) for c in _json.get('statusCount', list())],
        )


class PipelineStats:
    def __init__(
            self,
            pipeline_counters: List[PipelineCounter],
            node_counters: List[NodeCounters],
            pipeline_averages: PipelineAverages,
            node_averages: List[NodeAverages]
    ):
        self.pipeline_counters = pipeline_counters
        self.node_counters = node_counters
        self.pipeline_averages = pipeline_averages
        self.node_averages = node_averages

    @classmethod
    def from_json(cls, _json: dict):
        return cls(
            pipeline_counters=[PipelineCounter(**c) for c in _json.get('pipelineExecutionCounters', list())],
            node_counters=[NodeCounters.from_json(_json=c) for c in _json.get('nodeExecutionsCounters', list())],
            pipeline_averages=PipelineAverages.from_json(_json.get('pipelineExecutionStatistics', None)),
            node_averages=[NodeAverages.from_json(_json=c) for c in _json.get('nodeExecutionStatistics', list())]
        )


@attr.s
class Pipeline(entities.BaseEntity):
    """
    Pipeline object
    """
    # platform
    id = attr.ib()
    name = attr.ib()
    creator = attr.ib()
    org_id = attr.ib()
    connections = attr.ib()
    settings = attr.ib(type=PipelineSettings)
    variables = attr.ib(type=List[Variable])

    status = attr.ib(type=CompositionStatus)

    # name change
    created_at = attr.ib()
    updated_at = attr.ib(repr=False)
    start_nodes = attr.ib()
    project_id = attr.ib()
    composition_id = attr.ib()
    url = attr.ib()
    preview = attr.ib()
    description = attr.ib()
    revisions = attr.ib()

    # sdk
    _project = attr.ib(repr=False)
    _client_api = attr.ib(type=ApiClient, repr=False)
    _original_settings = attr.ib(repr=False, type=PipelineSettings)
    _original_variables = attr.ib(repr=False, type=List[Variable])
    _repositories = attr.ib(repr=False)

    updated_by = attr.ib(default=None)

    @staticmethod
    def _protected_from_json(_json, client_api, project=None, is_fetched=True):
        """
        Same as from_json but with try-except to catch if error
        :param _json: platform json
        :param client_api: ApiClient entity
        :param dtlpy.entities.project.Project project: entity
        :param is_fetched: is Entity fetched from Platform
        :return:
        """
        try:
            pipeline = Pipeline.from_json(
                _json=_json,
                client_api=client_api,
                project=project,
                is_fetched=is_fetched
            )
            status = True
        except Exception:
            pipeline = traceback.format_exc()
            status = False
        return status, pipeline

    @classmethod
    def from_json(cls, _json, client_api, project=None, is_fetched=True):
        """
        Turn platform representation of pipeline into a pipeline entity

        :param dict _json: platform representation of package
        :param dl.ApiClient client_api: ApiClient entity
        :param dtlpy.entities.project.Project project: entity
        :param bool is_fetched: is Entity fetched from Platform
        :return: Pipeline entity
        :rtype: dtlpy.entities.pipeline.Pipeline
        """
        if project is not None:
            if project.id != _json.get('projectId', None):
                logger.warning('Pipeline has been fetched from a project that is not belong to it')
                project = None

        connections = [PipelineConnection.from_json(_json=con) for con in _json.get('connections', list())]
        json_variables = _json.get('variables', None) or list()
        variables = list()
        if json_variables:
            copy_json_variables = copy.deepcopy(json_variables)
            variables = [Variable.from_json(_json=v) for v in copy_json_variables]

        settings = PipelineSettings.from_json(_json=_json.get('settings', dict()))
        inst = cls(
            created_at=_json.get('createdAt', None),
            updated_at=_json.get('updatedAt', None),
            project_id=_json.get('projectId', None),
            org_id=_json.get('orgId', None),
            composition_id=_json.get('compositionId', None),
            creator=_json.get('creator', None),
            client_api=client_api,
            name=_json.get('name', None),
            project=project,
            id=_json.get('id', None),
            connections=connections,
            start_nodes=_json.get('startNodes', None),
            url=_json.get('url', None),
            preview=_json.get('preview', None),
            description=_json.get('description', None),
            revisions=_json.get('revisions', None),
            settings=settings,
            variables=variables,
            status=_json.get('status', None),
            original_settings=settings,
            original_variables=json_variables,
            updated_by=_json.get('updatedBy', None),
        )
        for node in _json.get('nodes', list()):
            inst.nodes.add(node=cls.pipeline_node(node))
        inst.is_fetched = is_fetched
        return inst

    @classmethod
    def pipeline_node(cls, _json):
        node_type = _json.get('type')
        if node_type == 'task':
            return TaskNode.from_json(_json)
        elif node_type == 'code':
            return CodeNode.from_json(_json)
        elif node_type == 'function':
            return FunctionNode.from_json(_json)
        elif node_type == 'storage':
            return DatasetNode.from_json(_json)
        else:
            return PipelineNode.from_json(_json)

    def settings_changed(self) -> bool:
        return self.settings.to_json() != self._original_settings.to_json()

    def variables_changed(self) -> bool:
        new_vars = [var.to_json() for var in self.variables]
        old_vars = self._original_variables or list()
        return new_vars != old_vars

    def to_json(self):
        """
        Turn Package entity into a platform representation of Package

        :return: platform json of package
        :rtype: dict
        """
        _json = attr.asdict(self,
                            filter=attr.filters.exclude(attr.fields(Pipeline)._project,
                                                        attr.fields(Pipeline)._repositories,
                                                        attr.fields(Pipeline)._client_api,
                                                        attr.fields(Pipeline).project_id,
                                                        attr.fields(Pipeline).org_id,
                                                        attr.fields(Pipeline).connections,
                                                        attr.fields(Pipeline).created_at,
                                                        attr.fields(Pipeline).updated_at,
                                                        attr.fields(Pipeline).start_nodes,
                                                        attr.fields(Pipeline).project_id,
                                                        attr.fields(Pipeline).composition_id,
                                                        attr.fields(Pipeline).url,
                                                        attr.fields(Pipeline).preview,
                                                        attr.fields(Pipeline).description,
                                                        attr.fields(Pipeline).revisions,
                                                        attr.fields(Pipeline).settings,
                                                        attr.fields(Pipeline).variables,
                                                        attr.fields(Pipeline)._original_settings,
                                                        attr.fields(Pipeline)._original_variables,
                                                        attr.fields(Pipeline).updated_by,
                                                        ))

        _json['projectId'] = self.project_id
        _json['createdAt'] = self.created_at
        _json['updatedAt'] = self.updated_at
        _json['compositionId'] = self.composition_id
        _json['startNodes'] = self.start_nodes
        _json['orgId'] = self.org_id
        _json['nodes'] = [node.to_json() for node in self.nodes]
        _json['connections'] = [con.to_json() for con in self.connections]
        if self.variables:
            _json['variables'] = [v.to_json() for v in self.variables]
        _json['url'] = self.url

        settings_json = self.settings.to_json()
        if settings_json:
            _json['settings'] = settings_json

        if self.preview is not None:
            _json['preview'] = self.preview
        if self.description is not None:
            _json['description'] = self.description
        if self.revisions is not None:
            _json['revisions'] = self.revisions
        if self.updated_by is not None:
            _json['updatedBy'] = self.updated_by

        return _json

    #########
    # Props #
    #########

    @property
    def platform_url(self):
        return self._client_api._get_resource_url("projects/{}/pipelines/{}".format(self.project_id, self.id))

    @property
    def project(self):
        if self._project is None:
            self._project = self.projects.get(project_id=self.project_id, fetch=None)
        assert isinstance(self._project, entities.Project)
        return self._project

    ################
    # repositories #
    ################
    @_repositories.default
    def set_repositories(self):
        reps = namedtuple('repositories',
                          field_names=['projects', 'pipelines', 'pipeline_executions', 'triggers', 'nodes'])

        r = reps(
            projects=repositories.Projects(client_api=self._client_api),
            pipelines=repositories.Pipelines(client_api=self._client_api, project=self._project),
            pipeline_executions=repositories.PipelineExecutions(
                client_api=self._client_api, project=self._project, pipeline=self
            ),
            triggers=repositories.Triggers(client_api=self._client_api, pipeline=self),
            nodes=repositories.Nodes(client_api=self._client_api, pipeline=self)
        )
        return r

    @property
    def projects(self):
        assert isinstance(self._repositories.projects, repositories.Projects)
        return self._repositories.projects

    @property
    def triggers(self):
        assert isinstance(self._repositories.triggers, repositories.Triggers)
        return self._repositories.triggers

    @property
    def nodes(self):
        assert isinstance(self._repositories.nodes, repositories.Nodes)
        return self._repositories.nodes

    @property
    def pipelines(self):
        assert isinstance(self._repositories.pipelines, repositories.Pipelines)
        return self._repositories.pipelines

    @property
    def pipeline_executions(self):
        assert isinstance(self._repositories.pipeline_executions, repositories.PipelineExecutions)
        return self._repositories.pipeline_executions

    ###########
    # methods #
    ###########
    def update(self):
        """
        Update pipeline changes to platform

        :return: pipeline entity
        """
        return self.pipelines.update(pipeline=self)

    def delete(self):
        """
        Delete pipeline object

        :return: True
        """
        return self.pipelines.delete(pipeline=self)

    def open_in_web(self):
        """
        Open the pipeline in web platform

        :return:
        """
        self._client_api._open_in_web(url=self.platform_url)

    def install(self, resume_option: PipelineResumeOption = None):
        """
        install pipeline

        :return: Composition entity
        """
        return self.pipelines.install(pipeline=self, resume_option=resume_option)

    def pause(self, keep_triggers_active: bool = None):
        """
        pause pipeline

        :return: Composition entity
        """
        return self.pipelines.pause(pipeline=self, keep_triggers_active=keep_triggers_active)

    def execute(self, execution_input=None, node_id: str = None):
        """
        execute a pipeline and return to execute

        :param execution_input: list of the dl.FunctionIO or dict of pipeline input - example {'item': 'item_id'}
        :param str node_id: node id to execute
        :return: entities.PipelineExecution object
        """
        execution = self.pipeline_executions.create(
            pipeline_id=self.id,
            execution_input=execution_input,
            node_id=node_id
        )
        return execution

    def test(self, execution_input=None):
        """
        Execute a pipeline in test mode and return the pipeline execution as an object.

        :param execution_input: list of the dl.FunctionIO or dict of pipeline input - example {'item': 'item_id'}
        :return: entities.PipelineExecution object
        """
        execution = self.pipelines.test(
            pipeline=self,
            pipeline_id=self.id,
            execution_input=execution_input,
        )
        return execution

    def execute_batch(
            self,
            filters,
            execution_inputs=None,
            wait=True,
            node_id: str = None
    ):
        """
        execute a pipeline and return to execute

        :param execution_inputs: list of the dl.FunctionIO or dict of pipeline input - example {'item': 'item_id'}, that represent the extra inputs of the function
        :param filters: Filters entity for a filtering before execute
        :param bool wait: wait until create task finish
        :param str node_id: node id to execute
        :return: entities.PipelineExecution object

        **Example**:

        .. code-block:: python

            command = pipeline.execute_batch(
                        execution_inputs=dl.FunctionIO(type=dl.PackageInputType.STRING, value='test', name='string'),
                        filters=dl.Filters(field='dir', values='/test', context={'datasets': [dataset.id]))
        """
        command = self.pipeline_executions.create_batch(
            pipeline_id=self.id,
            execution_inputs=execution_inputs,
            filters=filters,
            wait=wait,
            node_id=node_id
        )
        return command

    def reset(self, stop_if_running: bool = False):
        """
        Resets pipeline counters

        :param bool stop_if_running: If the pipeline is installed it will stop the pipeline and reset the counters.
        :return: bool
        """
        return self.pipelines.reset(pipeline_id=self.id, stop_if_running=stop_if_running)

    def stats(self):
        """
        Get pipeline counters

        :return: PipelineStats
        :rtype: dtlpy.entities.pipeline.PipelineStats
        """
        return self.pipelines.stats(pipeline_id=self.id)

    def set_start_node(self, node: PipelineNode):
        """
        Set the start node of the pipeline

        :param PipelineNode node: node to be the start node
        """
        connections = [connection for connection in self.connections if connection.target.node_id == node.node_id]
        if connections:
            raise Exception(
                'Connections cannot be added to Pipeline start-node. To add a connection, please reposition the start sign')
        if self.start_nodes:
            for pipe_node in self.start_nodes:
                if pipe_node['type'] == 'root':
                    pipe_node['nodeId'] = node.node_id
        else:
            self.start_nodes = [{"nodeId": node.node_id,
                                 "type": "root", }]

    def update_variables_values(self, **kwargs):
        """
        Update pipeline variables values for the given keyword arguments.

        **Example**:

        .. code-block:: python
            pipeline.update_variables_values(
                dataset=dataset.id,
                model=model.id,
                threshold=0.9
            )
            pipeline.update()
        """
        keys = kwargs.keys()
        for variable in self.variables:
            if variable.name in keys:
                variable.value = kwargs[variable.name]

    def validate(self):
        """
        Validate the pipeline configuration.

        **prerequisites**: You must be an *owner* or *developer* to use this method.

        :return: Validation result
        :rtype: dict

        **Example**:

        .. code-block:: python

            validation_result = pipeline.validate()
        """
        return self.pipelines.validate(pipeline_json=self.to_json())



import logging
import typing
from enum import Enum

from .. import entities, utilities

logger = logging.getLogger(name='dtlpy')


class AppModule(entities.PackageModule, utilities.BaseServiceRunner):
    def __init__(self, name, description):
        super(AppModule, self).__init__()
        self.name = name
        self.description = description

    def init(self):
        """
        Default init function
        :return:
        """
        ...

    ###############
    # Decorators #
    ###############
    def set_init(self, inputs=None):
        def decorator(func: typing.Callable):
            setattr(self, 'init', func)
            setattr(self, func.__name__, func)
            init_inputs, _ = self._function_io_from_def(func=func, inputs=inputs, outputs=None)
            self.init_inputs = AppModule._parse_io(io_list=init_inputs)
            return func

        return decorator

    def add_function(self, display_name=None, inputs=None, outputs=None):
        """
        Add a function to the instance to be able to run in the FaaS.

        :param display_name: Display name of the function
        :param inputs: function inputs variable definitions. dictionary with strings for name and type {name: type}
        :param outputs: function output variable definitions. dictionary with strings for name and type {name: type}
        :return:
        """

        def decorator(func: typing.Callable):
            if display_name is None:
                d_name = func.__name__
            else:
                d_name = display_name
            module_inputs, module_outputs = self._function_io_from_def(func=func, inputs=inputs, outputs=outputs)
            defs = {"name": func.__name__,
                    "displayName": d_name,
                    "input": AppModule._parse_io(io_list=module_inputs),
                    "output": AppModule._parse_io(io_list=module_outputs)}
            func.__dtlpy__ = defs
            setattr(self, func.__name__, func)
            self.functions.append(entities.PackageFunction.from_json(defs))
            return func

        return decorator

    #########
    # Utils #
    #########
    @staticmethod
    def _function_io_from_def(func, inputs, outputs):
        input_types = typing.get_type_hints(func)
        hint_outputs = input_types.pop('return', None)

        if outputs is None:
            if hint_outputs is not None:
                func_outputs = {'output': hint_outputs}
            else:
                func_outputs = dict()
        else:
            func_outputs = outputs
        if inputs is None:
            func_inputs = {name: t for name, t in input_types.items()}
        else:
            func_inputs = inputs
        return func_inputs, func_outputs

    @staticmethod
    def _function_type_mapping(io_type):
        mapping = {'str': entities.PackageInputType.STRING}
        if io_type in mapping:
            return mapping[io_type]
        else:
            return io_type

    @staticmethod
    def _parse_io(io_list: dict):
        output = list()
        if io_list is not None:
            for io_name, io_type in io_list.items():
                if isinstance(io_type, Enum):
                    io_type = io_type.name
                if hasattr(io_type, '__args__'):
                    io_type = io_type.__args__[0]
                if isinstance(io_type, type):
                    io_type = io_type.__name__
                else:
                    io_type = str(io_type)
                output.append({"name": io_name,
                               "type": AppModule._function_type_mapping(str(io_type))})
        return output



import numpy as np

from . import BaseAnnotationDefinition


class Ellipse(BaseAnnotationDefinition):
    """
        Ellipse annotation object
    """
    type = "ellipse"

    def __init__(self, x, y, rx, ry, angle, label, attributes=None, description=None):
        super().__init__(description=description, attributes=attributes)
        self.label = label
        self.angle = angle
        self.x = x
        self.y = y
        self.rx = rx
        self.ry = ry

    @property
    def geo(self):
        return np.asarray([[self.x, self.y],
                           [self.rx, self.ry],
                           [self.angle, self.rad]])

    @property
    def rad(self):
        return np.deg2rad(self.angle)

    @property
    def left(self):
        return self.x - np.sqrt(np.power(self.rx, 2) * np.power(np.cos(-self.rad), 2)
                                + np.power(self.ry, 2) * np.power(np.sin(-self.rad), 2))

    @property
    def top(self):
        return self.y - np.sqrt(np.power(self.rx, 2) * np.power(np.sin(-self.rad), 2)
                                + np.power(self.ry, 2) * np.power(np.cos(-self.rad), 2))

    @property
    def right(self):
        return self.x + np.sqrt(np.power(self.rx, 2) * np.power(np.cos(-self.rad), 2)
                                + np.power(self.ry, 2) * np.power(np.sin(-self.rad), 2))

    @property
    def bottom(self):
        return self.y + np.sqrt(np.power(self.rx, 2) * np.power(np.sin(-self.rad), 2)
                                + np.power(self.ry, 2) * np.power(np.cos(-self.rad), 2))

    def show(self, image, thickness, with_text, height, width, annotation_format, color, alpha=1):
        """
        Show annotation as ndarray
        :param image: empty or image to draw on
        :param thickness:
        :param with_text: not required
        :param height: item height
        :param width: item width
        :param annotation_format: options: list(dl.ViewAnnotationOptions)
        :param color: color
        :param alpha: opacity value [0 1], default 1
        :return: ndarray
        """
        try:
            import cv2
        except (ImportError, ModuleNotFoundError):
            self.logger.error(
                'Import Error! Cant import cv2. Annotations operations will be limited. import manually and fix errors')
            raise

        if thickness is None:
            thickness = 2

        # draw annotation
        image = cv2.ellipse(
            image,
            center=(int(np.round(self.x)), int(np.round(self.y))),
            axes=(int(np.round(self.rx)), int(np.round(self.ry))),
            angle=self.angle,
            startAngle=0,
            endAngle=360,
            color=color,
            thickness=thickness,
            lineType=-1,
        )
        if with_text:
            image = self.add_text_to_image(image=image, annotation=self)
        return image

    def to_coordinates(self, color):
        return {'angle': float(self.angle),
                'center': {'x': float(self.x),
                           'y': float(self.y),
                           'z': 0},
                'rx': float(self.rx),
                'ry': float(self.ry)}

    @staticmethod
    def from_coordinates(coordinates):
        angle = coordinates["angle"]
        x = coordinates["center"]["x"]
        y = coordinates["center"]["y"]
        rx = coordinates["rx"]
        ry = coordinates["ry"]
        return x, y, rx, ry, angle

    @classmethod
    def from_json(cls, _json):
        if "coordinates" in _json:
            x, y, rx, ry, angle = cls.from_coordinates(_json["coordinates"])
        elif "data" in _json:
            x, y, rx, ry, angle = cls.from_coordinates(_json["data"])
        else:
            raise ValueError('can not find "coordinates" or "data" in annotation. id: {}'.format(_json["id"]))

        return cls(
            angle=angle,
            x=x,
            y=y,
            rx=rx,
            ry=ry,
            label=_json["label"],
            attributes=_json.get("attributes", None),
        )



import logging
from urllib.parse import quote

from .. import entities, miscellaneous, exceptions, _api_reference
from ..services.api_client import ApiClient

logger = logging.getLogger(name='dtlpy')


class Projects:
    """
    Projects Repository

    The Projects class allows the user to manage projects and their properties.

    For more information on Projects see the `Dataloop documentation <https://dataloop.ai/docs/project#>`_.
    """

    def __init__(self, client_api: ApiClient, org=None):
        self._client_api = client_api
        self._org = org

    def __get_from_cache(self) -> entities.Project:
        project = self._client_api.state_io.get('project')
        if project is not None:
            project = entities.Project.from_json(_json=project, client_api=self._client_api)
        return project

    def __get_by_id(self, project_id: str, log_error: bool) -> entities.Project:
        """
        :param project_id:
        """
        success, response = self._client_api.gen_request(
            req_type='get',
            path='/projects/{}'.format(project_id),
            log_error=log_error
        )

        try:
            response_json = response.json()
        except Exception:
            try:
                logger.exception('Failed to parse response content: {}'.format(response.text))
            except Exception:
                logger.exception('Failed to print response content')
            raise

        if success:
            project = entities.Project.from_json(
                client_api=self._client_api,
                _json=response_json
            )
        else:
            # raise PlatformException(response)
            # TODO because of a bug in gate wrong error is returned so for now manually raise not found
            raise exceptions.PlatformException(error="404", message="Project not found")
        return project

    def __get_by_name(self, project_name: str):
        """
        :param project_name:
        """
        project_name = quote(project_name.encode("utf-8"))
        success, response = self._client_api.gen_request(req_type='get',
                                                         path='/projects/{}/name'.format(project_name))
        if success:
            projects = [entities.Project.from_json(client_api=self._client_api,
                                                   _json=project_json) for project_json in response.json()]
        else:
            # TODO because of a bug in gate wrong error is returned so for now manually raise not found
            raise exceptions.PlatformException(error="404", message="Project not found")
        return projects

    def __get_by_identifier(self, identifier=None) -> entities.Project:
        """
        :param identifier:
        """
        projects = self.list()
        projects_by_name = [project for project in projects if identifier in project.id or identifier in project.name]
        if len(projects_by_name) == 1:
            return projects_by_name[0]
        elif len(projects_by_name) > 1:
            raise Exception('Multiple projects with this name/identifier exist')
        else:
            raise Exception("Project not found")

    @property
    def platform_url(self):
        return self._client_api._get_resource_url("projects")

    def open_in_web(self,
                    project_name: str = None,
                    project_id: str = None,
                    project: entities.Project = None):
        """
        Open the project in our web platform.

        **Prerequisites**: All users can open a project in the web.

        :param str project_name: The Name of the project
        :param str project_id: The Id of the project
        :param dtlpy.entities.project.Project project: project object

        **Example**:

        .. code-block:: python

            dl.projects.open_in_web(project_id='project_id')
        """
        if project_name is not None:
            project = self.get(project_name=project_name)
        if project is not None:
            project.open_in_web()
        elif project_id is not None:
            self._client_api._open_in_web(url=self.platform_url + '/' + str(project_id))
        else:
            self._client_api._open_in_web(url=self.platform_url)

    def checkout(self,
                 identifier: str = None,
                 project_name: str = None,
                 project_id: str = None,
                 project: entities.Project = None):
        """
        Checkout (switch) to a project to work on.

        **Prerequisites**: All users can open a project in the web.

        You must provide at least ONE of the following params: project_id, project_name.

        :param str identifier: project name or partial id that you wish to switch
        :param str project_name: The Name of the project
        :param str project_id: The Id of the project
        :param dtlpy.entities.project.Project project: project object

        **Example**:

        .. code-block:: python

            dl.projects.checkout(project_id='project_id')
        """
        if project is None:
            if project_id is not None or project_name is not None:
                project = self.get(project_id=project_id, project_name=project_name)
            elif identifier is not None:
                project = self.__get_by_identifier(identifier=identifier)
            else:
                raise exceptions.PlatformException(error='400',
                                                   message='Must provide partial/full id/name to checkout')
        self._client_api.state_io.put('project', project.to_json())
        logger.info('Checked out to project {}'.format(project.name))

    def _send_mail(self, project_id: str, send_to: str, title: str, content: str) -> bool:
        if project_id:
            url = '/projects/{}/mail'.format(project_id)
        else:
            url = '/outbox'
        assert isinstance(title, str)
        assert isinstance(content, str)
        if self._client_api.token is not None:
            sender = self._client_api.info()['user_email']
        else:
            raise exceptions.PlatformException('600', 'Token expired please log in')

        payload = {
            'to': send_to,
            'from': sender,
            'subject': title,
            'body': content
        }

        success, response = self._client_api.gen_request(req_type='post',
                                                         path=url,
                                                         json_req=payload)

        if not success:
            raise exceptions.PlatformException(response)
        return True

    @_api_reference.add(path='/projects/{projectId}/members/{userId}', method='post')
    def add_member(self, email: str, project_id: str, role: entities.MemberRole = entities.MemberRole.DEVELOPER):
        """
        Add a member to the project.

        **Prerequisites**: You must be in the role of an *owner* to add a member to a project.

        :param str email: member email
        :param str project_id: The Id of the project
        :param role: The required role for the user. Use the enum dl.MemberRole
        :return: dict that represent the user
        :rtype: dict

        **Example**:

        .. code-block:: python

            user_json = dl.projects.add_member(project_id='project_id', email='user@dataloop.ai', role=dl.MemberRole.DEVELOPER)
        """
        url_path = '/projects/{}/members/{}'.format(project_id, email)
        payload = dict(role=role)

        if role not in list(entities.MemberRole):
            raise ValueError('Unknown role {!r}, role must be one of: {}'.format(role,
                                                                                 ', '.join(list(entities.MemberRole))))

        success, response = self._client_api.gen_request(req_type='post',
                                                         path=url_path,
                                                         json_req=payload)
        if not success:
            raise exceptions.PlatformException(response)

        return response.json()

    @_api_reference.add(path='/projects/{projectId}/members/{userId}', method='patch')
    def update_member(self, email: str, project_id: str, role: entities.MemberRole = entities.MemberRole.DEVELOPER):
        """
        Update member's information/details in the project.

        **Prerequisites**: You must be in the role of an *owner* to update a member.

        :param str email: member email
        :param str project_id: The Id of the project
        :param role: The required role for the user. Use the enum dl.MemberRole
        :return: dict that represent the user
        :rtype: dict

        **Example**:

        .. code-block:: python

            user_json = = dl.projects.update_member(project_id='project_id', email='user@dataloop.ai', role=dl.MemberRole.DEVELOPER)
        """
        url_path = '/projects/{}/members/{}'.format(project_id, email)
        payload = dict(role=role)

        if role not in list(entities.MemberRole):
            raise ValueError('Unknown role {!r}, role must be one of: {}'.format(role,
                                                                                 ', '.join(list(entities.MemberRole))))

        success, response = self._client_api.gen_request(req_type='patch',
                                                         path=url_path,
                                                         json_req=payload)
        if not success:
            raise exceptions.PlatformException(response)

        return response.json()

    @_api_reference.add(path='/projects/{projectId}/members/{userId}', method='delete')
    def remove_member(self, email: str, project_id: str):
        """
        Remove a member from the project.

        **Prerequisites**: You must be in the role of an *owner* to delete a member from a project.

        :param str email: member email
        :param str project_id: The Id of the project
        :return: dict that represents the user
        :rtype: dict

        **Example**:

        .. code-block:: python

            user_json = dl.projects.remove_member(project_id='project_id', email='user@dataloop.ai')
        """
        url_path = '/projects/{}/members/{}'.format(project_id, email)
        success, response = self._client_api.gen_request(req_type='delete',
                                                         path=url_path)
        if not success:
            raise exceptions.PlatformException(response)

        return response.json()

    @_api_reference.add(path='/projects/{projectId}/members', method='get')
    def list_members(self, project: entities.Project, role: entities.MemberRole = None):
        """
        Get a list of the project members.

        **Prerequisites**: You must be in the role of an *owner* to list project members.

        :param dtlpy.entities.project.Project project: Project object
        :param role: The required role for the user. Use the enum dl.MemberRole
        :return: list of the project members
        :rtype: list

        **Example**:

        .. code-block:: python

            users_jsons_list = dl.projects.list_members(project_id='project_id', role=dl.MemberRole.DEVELOPER)
        """
        url_path = '/projects/{}/members'.format(project.id)

        if role is not None and role not in list(entities.MemberRole):
            raise ValueError('Unknown role {!r}, role must be one of: {}'.format(role,
                                                                                 ', '.join(list(entities.MemberRole))))

        success, response = self._client_api.gen_request(req_type='get',
                                                         path=url_path)
        if not success:
            raise exceptions.PlatformException(response)

        members = miscellaneous.List(
            [entities.User.from_json(_json=user, client_api=self._client_api, project=project) for user in
             response.json()])

        if role is not None:
            members = [member for member in members if member.role == role]

        return members

    @_api_reference.add(path='/projects', method='get')
    def list(self) -> miscellaneous.List[entities.Project]:
        """
        Get the user's project list

        **Prerequisites**: You must be a **superuser** to list all users' projects.

        :return: List of Project objects

        **Example**:

        .. code-block:: python

            projects = dl.projects.list()
        """
        if self._org is None:
            url_path = '/projects'
        else:
            url_path = '/orgs/{}/projects'.format(self._org.id)
        success, response = self._client_api.gen_request(req_type='get',
                                                         path=url_path)

        if success:
            pool = self._client_api.thread_pools(pool_name='entity.create')
            projects_json = response.json()
            jobs = [None for _ in range(len(projects_json))]
            # return triggers list
            for i_project, project in enumerate(projects_json):
                jobs[i_project] = pool.submit(entities.Project._protected_from_json,
                                              **{'client_api': self._client_api,
                                                 '_json': project})

            # get all results
            results = [j.result() for j in jobs]
            # log errors
            _ = [logger.warning(r[1]) for r in results if r[0] is False]
            # return good jobs
            projects = miscellaneous.List([r[1] for r in results if r[0] is True])
        else:
            logger.error('Platform error getting projects')
            raise exceptions.PlatformException(response)
        return projects

    @_api_reference.add(path='/projects/{projectId}', method='get')
    def get(self,
            project_name: str = None,
            project_id: str = None,
            checkout: bool = False,
            fetch: bool = None,
            log_error=True) -> entities.Project:
        """
        Get a Project object.

        **Prerequisites**: You must be in the role of an *owner* to get a project object.

        You must check out to a project or provide at least one of the following params: project_id, project_name

        :param str project_name: optional - search by name
        :param str project_id: optional - search by id
        :param bool checkout: set the project as a default project object (cookies)
        :param bool fetch: optional - fetch entity from platform (True), default taken from cookie
        :param bool log_error: optional - show the logs errors
        :return: Project object
        :rtype: dtlpy.entities.project.Project

        **Example**:

        .. code-block:: python

            project = dl.projects.get(project_id='project_id')
        """
        if fetch is None:
            fetch = self._client_api.fetch_entities

        if project_id is None and project_name is None:
            project = self.__get_from_cache()
            if project is None:
                raise exceptions.PlatformException(
                    error='400',
                    message='No checked-out Project was found. You must checkout to a project or provide an identifier in inputs')
        elif fetch:
            if project_id is not None:
                if not isinstance(project_id, str):
                    raise exceptions.PlatformException(
                        error='400',
                        message='project_id must be strings')

                project = self.__get_by_id(project_id, log_error=log_error)
                # verify input project name is same as the given id
                if project_name is not None and project.name != project_name:
                    logger.warning(
                        "Mismatch found in projects.get: project_name is different then project.name:"
                        " {!r} != {!r}".format(
                            project_name,
                            project.name))
            elif project_name is not None:
                if not isinstance(project_name, str):
                    raise exceptions.PlatformException(
                        error='400',
                        message='project_name must be strings')

                projects = self.__get_by_name(project_name)
                if len(projects) > 1:
                    # more than one matching project
                    raise exceptions.PlatformException(
                        error='404',
                        message='More than one project with same name. Please "get" by id')
                else:
                    project = projects[0]
            else:
                raise exceptions.PlatformException(
                    error='404',
                    message='No input and no checked-out found')
        else:
            project = entities.Project.from_json(_json={'id': project_id,
                                                        'name': project_name},
                                                 client_api=self._client_api,
                                                 is_fetched=False)
        assert isinstance(project, entities.Project)
        if checkout:
            self.checkout(project=project)
        return project

    @_api_reference.add(path='/projects/{projectId}', method='delete')
    def delete(self,
               project_name: str = None,
               project_id: str = None,
               sure: bool = False,
               really: bool = False) -> bool:
        """
        Delete a project forever!

        **Prerequisites**: You must be in the role of an *owner* to delete a project.

        :param str project_name: optional - search by name
        :param str project_id: optional - search by id
        :param bool sure: Are you sure you want to delete?
        :param bool really: Really really sure?
        :return: True if success, error if not
        :rtype: bool

        **Example**:

        .. code-block:: python

            is_deleted = dl.projects.delete(project_id='project_id', sure=True, really=True)
        """
        if sure and really:
            if project_id is None:
                project = self.get(project_name=project_name)
                project_id = project.id
            success, response = self._client_api.gen_request(req_type='delete',
                                                             path='/projects/{}'.format(project_id))
            if not success:
                raise exceptions.PlatformException(response)
            logger.info('Project id {} deleted successfully'.format(project_id))
            return True
        else:
            raise exceptions.PlatformException(
                error='403',
                message='Cant delete project from SDK. Please login to platform to delete')

    @_api_reference.add(path='/projects/{projectId}', method='patch')
    def update(self,
               project: entities.Project,
               system_metadata: bool = False) -> entities.Project:
        """
        Update a project information (e.g., name, member roles, etc.).

        **Prerequisites**: You must be in the role of an *owner* to add a member to a project.

        :param dtlpy.entities.project.Project project: project object
        :param bool system_metadata: optional - True, if you want to change metadata system
        :return: Project object
        :rtype: dtlpy.entities.project.Project

        **Example**:

        .. code-block:: python

            project = dl.projects.delete(project='project_entity')
        """
        url_path = '/projects/{}'.format(project.id)
        if system_metadata:
            url_path += '?system=true'
        success, response = self._client_api.gen_request(req_type='patch',
                                                         path=url_path,
                                                         json_req=project.to_json())
        if success:
            return project
        else:
            raise exceptions.PlatformException(response)

    @_api_reference.add(path='/projects', method='post')
    def create(self,
               project_name: str,
               checkout: bool = False) -> entities.Project:
        """
        Create a new project.

        **Prerequisites**: Any user can create a project.

        :param str project_name: The Name of the project
        :param bool checkout: set the project as a default project object (cookies)
        :return: Project object
        :rtype: dtlpy.entities.project.Project

        **Example**:

        .. code-block:: python

            project = dl.projects.create(project_name='project_name')
        """
        payload = {'name': project_name}
        success, response = self._client_api.gen_request(req_type='post',
                                                         path='/projects',
                                                         data=payload)
        if success:
            project = entities.Project.from_json(client_api=self._client_api,
                                                 _json=response.json())
        else:
            raise exceptions.PlatformException(response)
        assert isinstance(project, entities.Project)
        if checkout:
            self.checkout(project=project)
        return project



from urllib.parse import urlsplit, urlunsplit
import base64
import requests
import logging
import json
import jwt
import os

logger = logging.getLogger(name='dtlpy')


def login_m2m(api_client, email, password, client_id=None, client_secret=None, force=False):
    return login_secret(api_client=api_client,
                        email=email,
                        password=password,
                        client_id=client_id,
                        client_secret=client_secret,
                        force=force)


def login_secret(api_client, email, password, client_id, client_secret=None, force=False):
    """
    Login with email and password from environment variables
    :param api_client: ApiClient instance
    :param email: user email. if already logged in with same user - login will NOT happen. see "force"
    :param password: user password
    :param client_id: DEPRECATED
    :param client_secret: DEPRECATED
    :param force: force login. in case login with same user but want to get a new JWT
    :return:
    """
    # TODO add deprecation warning to client_id
    # check if already logged in with SAME email
    if api_client.token is not None or api_client.token == '':
        logged_email = api_client.info()['user_email']
        if logged_email == email and \
            not api_client.token_expired() and \
            not force:
            return True
        else:
            logger.debug('{}'.format('Cant decode token. Force login is used'))

    logger.info('[Start] Login Secret')
    env_params = api_client.environments[api_client.environment]
    # need to login
    payload = {'username': email,
               'password': password,
               'type': 'user_credentials'
               }
    headers = {'content-type': 'application/json'}
    if 'gate_url' not in env_params:
        env_params['gate_url'] = gate_url_from_host(environment=api_client.environment)
        api_client.environments[api_client.environment] = env_params
    token_url = env_params['gate_url'] + "/token?default"
    resp = requests.request("POST",
                            token_url,
                            data=json.dumps(payload),
                            headers=headers,
                            verify=env_params.get('verify_ssl', True))
    if not resp.ok:
        logout(api_client=api_client)
        api_client.print_bad_response(resp)
        return False
    else:
        response_dict = resp.json()
        api_client.token = response_dict['id_token']  # this will also set the refresh_token to None
        if 'refresh_token' in response_dict:
            api_client.refresh_token = response_dict['refresh_token']

        # set new client id for refresh
        logged_email = api_client.info()['user_email']

        if not logged_email == 'null':
            logger.info(f'[Done] Login Secret. User: {logged_email}')
        else:
            logger.info(f'[Done] Login Secret. User: {email}')
    return True


def logout(api_client):
    """
    remove JWT from cookie
    """
    api_client.token = None
    api_client.refresh_token = None
    return True


def login_html():
    try:
        location = os.path.dirname(os.path.realpath(__file__))
    except NameError:
        location = './dtlpy/services'
    filename = os.path.join(location, '..', 'assets', 'lock_open.png')

    if os.path.isfile(filename):

        with open(filename, 'rb') as f:
            image = f.read()

        html = (
            "        <!doctype html>\n"
            "        <html>\n"
            "        <head>\n"
            "            <style>\n"
            "                body {{\n"
            "                    background-color: #F7F7F9 !important;\n"
            "                    display: flex;\n"
            "                    justify-content: center;\n"
            "                    align-items: center;\n"
            "                    height: 100vh;\n"
            "                    width: 100vw;\n"
            "                    margin: 0;\n"
            "                }}\n"
            "                img {{\n"
            "                    display: block;\n"
            "                    max-width: 100%;\n"
            "                    max-height: 100%;\n"
            "                    margin: auto;\n"
            "                }}\n"
            "            </style>\n"
            "        </head>\n"
            "        <body>\n"
            "            <img src='data:image/png;base64,{image}'>\n"
            "        </body>\n"
            "        </html>\n"
        ).format(image=base64.b64encode(image).decode())
    else:
        html = "<!doctype html><html><body>Logged in successfully</body></html>"

    return html.encode('utf-8')


def login(api_client, auth0_url=None, audience=None, client_id=None, login_domain=None, callback_port=None):
    import webbrowser
    from http.server import BaseHTTPRequestHandler, HTTPServer
    from urllib.parse import urlparse, parse_qs
    logger.info('Logging in to Dataloop...')

    class LocalServer:

        class Handler(BaseHTTPRequestHandler):

            tokens_obtained = False
            id_token = None
            access_token = None
            refresh_token = None

            def log_message(self, format, *args):
                return

            def do_GET(self):
                parsed_path = urlparse(self.path)
                query = parse_qs(parsed_path.query)
                self.send_response(200)
                self.send_header('Content-type', 'text/html')
                self.end_headers()
                self.wfile.write(login_html())
                self.__class__.id_token = query['id_token'][0]
                self.__class__.access_token = query['access_token'][0]
                self.__class__.refresh_token = query['refresh_token'][0]
                self.__class__.tokens_obtained = True

        def __init__(self):
            self.port = callback_port if callback_port is not None else 3001
            self.server = HTTPServer(('', self.port), self.Handler)
            self.server.timeout = 60

        def process_request(self):
            self.server.handle_request()

            if self.Handler.tokens_obtained:
                return True, {
                    "id": self.Handler.id_token,
                    "access": self.Handler.access_token,
                    "refresh": self.Handler.refresh_token
                }
            else:
                return False, {}

        def local_endpoint(self):
            return "http://localhost:{}".format(self.port)

        def close(self):
            self.server.server_close()

    server = LocalServer()

    try:
        local_ep = server.local_endpoint()
        env_params = api_client.environments[api_client.environment]
        if 'gate_url' not in env_params:
            env_params['gate_url'] = gate_url_from_host(environment=api_client.environment)
            api_client.environments[api_client.environment] = env_params
        remote_ep = env_params['gate_url']
        login_page_url = "{}/login?callback={}".format(remote_ep, local_ep)
        if login_domain is not None:
            login_page_url = "{}&domain={}".format(login_page_url, login_domain)
        logger.info("Launching interactive login via {}".format(remote_ep))
        webbrowser.open(url=login_page_url, new=2, autoraise=True)

        success, tokens = server.process_request()

        if success:
            # oxsec-disable jwt-signature-disabled - Client-side SDK: signature verification disabled intentionally to extract claims for display; server validates on API calls
            decoded_jwt = jwt.decode(
                tokens['id'],
                options={
                    "verify_signature": False,
                    "verify_exp": False,
                    "verify_aud": False,
                    "verify_iss": False,
                }
            )

            if 'email' in decoded_jwt:
                logger.info('Logged in: {}'.format(decoded_jwt['email']))
            else:
                logger.info('Logged in: unknown user')

            api_client.token = tokens['id']
            api_client.refresh_token = tokens['refresh']

            return True
        else:
            logout(api_client=api_client)
            logger.error('Login failed: no tokens obtained')
            return False
    except Exception as err:
        logout(api_client=api_client)
        logger.exception('Login failed: error while getting token', err)
        return False
    finally:
        server.close()


def gate_url_from_host(environment):
    parsed = urlsplit(environment)
    return urlunsplit((parsed.scheme, parsed.netloc, '', '', ''))



from . import BaseAnnotationDefinition


class Subtitle(BaseAnnotationDefinition):
    """
        Subtitle annotation object
    """
    type = "subtitle"

    def __init__(self, text, label, attributes=None, description=None):
        super().__init__(description=description, attributes=attributes)
        self.text = text
        self.label = label

    def to_coordinates(self, color):
        return {"text": self.text}

    @staticmethod
    def from_coordinates(coordinates):
        return coordinates["text"]

    @classmethod
    def from_json(cls, _json):
        if "coordinates" in _json:
            text = cls.from_coordinates(coordinates=_json["coordinates"])
        elif "data" in _json:
            text = cls.from_coordinates(coordinates=_json["data"])
        else:
            raise ValueError('Bad json, "coordinates" or "data" not found')
        return cls(
            text=text,
            label=_json["label"],
            attributes=_json.get("attributes", None)
        )



import os
import validators
from dtlpy import PlatformException, entities


class BaseUploadElement:

    def __init__(self, all_upload_elements: dict):
        self.upload_item_element = all_upload_elements['upload_item_element']
        self.total_size = all_upload_elements['total_size']
        self.remote_name = all_upload_elements['remote_name']
        self.remote_path = all_upload_elements['remote_path']
        self.upload_annotations_element = all_upload_elements['upload_annotations_element']
        self.item_metadata = all_upload_elements['item_metadata']
        self.with_head_folder = all_upload_elements['with_head_folder']
        self.type = None
        self.buffer = None
        self.remote_filepath = None
        self.exists_in_remote = False
        self.checked_in_remote = False
        self.annotations_filepath = all_upload_elements['annotations_filepath']
        self.link_dataset_id = None
        self.root = all_upload_elements['root']
        self.filename = all_upload_elements['filename']
        self.export_version = all_upload_elements['export_version']
        self.item_description = all_upload_elements['item_description']
        self.driver_path = all_upload_elements['driver_path']


class BinaryUploadElement(BaseUploadElement):

    def __init__(self, all_upload_elements: dict):
        super().__init__(all_upload_elements)

        # binary element
        if not hasattr(self.upload_item_element, "name") and self.remote_name is None:
            raise PlatformException(
                error="400",
                message='upload element type was bytes. Must put attribute "name" on buffer (with file name) '
                        'when uploading buffers or providing param "remote_name"')
        if self.remote_name is None:
            self.remote_name = self.upload_item_element.name
        remote_filepath = self.remote_path + self.remote_name
        self.type = 'buffer'
        self.buffer = self.upload_item_element
        self.remote_filepath = remote_filepath
        self.annotations_filepath = self.upload_annotations_element


class DirUploadElement(BaseUploadElement):

    def __init__(self, all_upload_elements: dict):
        super().__init__(all_upload_elements)

        filepath = os.path.join(self.root, self.filename)  # get full image filepath
        # extract item's size
        self.total_size += os.path.getsize(filepath)
        # get annotations file
        if self.upload_annotations_element is not None:
            # change path to annotations
            annotations_filepath = filepath.replace(self.upload_item_element,
                                                    self.upload_annotations_element)
            # remove image extension
            if self.export_version == entities.ExportVersion.V1:
                annotations_filepath, _ = os.path.splitext(annotations_filepath)
            # add json extension
            annotations_filepath += ".json"
        else:
            annotations_filepath = None
        if self.with_head_folder:
            remote_filepath = self.remote_path + os.path.relpath(filepath, os.path.dirname(
                self.upload_item_element))
        else:
            remote_filepath = self.remote_path + os.path.relpath(filepath, self.upload_item_element)

        self.type = 'file'
        self.buffer = filepath
        self.remote_filepath = remote_filepath.replace("\\", "/")
        self.annotations_filepath = annotations_filepath


class ExternalItemUploadElement(BaseUploadElement):

    def __init__(self, all_upload_elements: dict):
        super().__init__(all_upload_elements)
        # self.upload_item_element = self.upload_item_element.split('//')[1]
        filepath = self.upload_item_element
        # extract item's size
        self.total_size += all_upload_elements['total_size']
        # determine item's remote base name
        if self.remote_name is None:
            self.remote_name = os.path.basename(filepath)
        # get annotations file
        if self.upload_annotations_element is not None:
            # change path to annotations
            annotations_filepath = filepath.replace(self.upload_item_element, self.upload_annotations_element)
            # remove image extension
            if self.export_version == entities.ExportVersion.V1:
                annotations_filepath, _ = os.path.splitext(annotations_filepath)
            # add json extension
            annotations_filepath += ".json"
        else:
            annotations_filepath = None
        # append to list
        if self.remote_path is None:
            split_dir = os.path.dirname(filepath).split('//')
            if len(split_dir) > 1:
                split_dir = split_dir[1] + '/'
            if self.driver_path == '/':
                self.driver_path = None
            if self.driver_path and split_dir.startswith(self.driver_path):
                self.remote_path = split_dir.replace(self.driver_path, '')
            else:
                self.remote_path = split_dir
        remote_filepath = self.remote_path + self.remote_name
        self.type = 'external_file'
        self.buffer = filepath
        self.remote_filepath = remote_filepath
        self.annotations_filepath = annotations_filepath


class FileUploadElement(BaseUploadElement):

    def __init__(self, all_upload_elements: dict):
        super().__init__(all_upload_elements)

        filepath = self.upload_item_element
        # extract item's size
        self.total_size += os.path.getsize(filepath)
        # determine item's remote base name
        if self.remote_name is None:
            self.remote_name = os.path.basename(filepath)
        # get annotations file
        if self.upload_annotations_element is not None:
            # change path to annotations
            annotations_filepath = filepath.replace(self.upload_item_element, self.upload_annotations_element)
            # remove image extension
            if self.export_version == entities.ExportVersion.V1:
                annotations_filepath, _ = os.path.splitext(annotations_filepath)
            # add json extension
            annotations_filepath += ".json"
        else:
            annotations_filepath = None
        # append to list
        remote_filepath = self.remote_path + self.remote_name
        self.type = 'file'
        self.buffer = filepath
        self.remote_filepath = remote_filepath
        self.annotations_filepath = annotations_filepath


class ItemLinkUploadElement(BaseUploadElement):

    def __init__(self, all_upload_elements: dict):
        super().__init__(all_upload_elements)

        link = entities.Link(ref=self.upload_item_element.id, type='id', dataset_id=self.upload_item_element.dataset_id,
                             name='{}_link.json'.format(self.upload_item_element.name))
        if self.remote_name is None:
            self.remote_name = link.name

        remote_filepath = '{}{}'.format(self.remote_path, self.remote_name)
        self.type = 'link'
        self.buffer = link
        self.remote_filepath = remote_filepath
        self.annotations_filepath = self.upload_annotations_element


class LinkUploadElement(BaseUploadElement):

    def __init__(self, all_upload_elements: dict):
        super().__init__(all_upload_elements)

        if self.remote_name is None:
            self.remote_name = self.upload_item_element.name

        remote_filepath = '{}{}_link.json'.format(self.remote_path, self.remote_name)
        self.type = 'link'
        self.buffer = self.upload_item_element
        self.remote_filepath = remote_filepath
        self.annotations_filepath = self.upload_annotations_element


class MultiViewUploadElement(BaseUploadElement):

    def __init__(self, all_upload_elements: dict):
        super().__init__(all_upload_elements)

        if self.remote_name:
            self.upload_item_element.name = self.remote_name

        if not self.upload_item_element.name.endswith('.json'):
            self.upload_item_element.name = '{}.json'.format(self.upload_item_element.name)

        remote_filepath = '{}{}'.format(self.remote_path, self.upload_item_element.name)
        self.type = 'collection'
        self.buffer = self.upload_item_element.to_bytes_io()
        self.remote_filepath = remote_filepath
        self.annotations_filepath = self.upload_annotations_element


class SimilarityUploadElement(BaseUploadElement):

    def __init__(self, all_upload_elements: dict):
        super().__init__(all_upload_elements)

        if self.remote_name:
            self.upload_item_element.name = self.remote_name
        if self.upload_item_element.name is None:
            self.upload_item_element.name = '{}_similarity.json'.format(self.upload_item_element.ref)

        if not self.upload_item_element.name.endswith('.json'):
            remote_filepath = '{}{}.json'.format(self.remote_path, self.upload_item_element.name)
        else:
            remote_filepath = '{}{}'.format(self.remote_path, self.upload_item_element.name)

        self.type = 'collection'
        self.buffer = self.upload_item_element.to_bytes_io()
        self.remote_filepath = remote_filepath
        self.annotations_filepath = self.upload_annotations_element


class PromptUploadElement(BaseUploadElement):

    def __init__(self, all_upload_elements: dict):
        super().__init__(all_upload_elements)

        if self.remote_name:
            self.upload_item_element.name = self.remote_name
        if self.upload_item_element.name is None:
            self.upload_item_element.name = '{}_prompt.json'.format(self.upload_item_element.ref)

        if not self.upload_item_element.name.endswith('.json'):
            remote_filepath = '{}{}.json'.format(self.remote_path, self.upload_item_element.name)
        else:
            remote_filepath = '{}{}'.format(self.remote_path, self.upload_item_element.name)

        self.type = 'prompt'
        self.buffer = self.upload_item_element.to_bytes_io()
        self.remote_filepath = remote_filepath
        self.annotations_filepath = self.upload_annotations_element


class UrlUploadElement(BaseUploadElement):

    def __init__(self, all_upload_elements: dict):
        super().__init__(all_upload_elements)

        # noinspection PyTypeChecker
        if self.remote_name is None:
            self.remote_name = str(self.upload_item_element.split('/')[-1]).split('?')[0]

        remote_filepath = self.remote_path + self.remote_name
        self.type = 'url'
        self.buffer = self.upload_item_element
        self.remote_filepath = remote_filepath
        self.annotations_filepath = self.upload_annotations_element



import os
import datetime
import json
import threading
from .. import services
from .. import exceptions
import logging
from ..caches import dl_cache

logger = logging.getLogger(name='dtlpy')
CHUNK = 200000


class Reporter:
    """
     ThreadPool Report summary
    """
    ITEMS_DOWNLOAD = 'downloader'
    ITEMS_UPLOAD = 'uploader'
    CONVERTER = 'converter'

    def __init__(self,
                 num_workers,
                 resource,
                 client_api,
                 print_error_logs=False,
                 output_entity=None,
                 no_output=False):
        self._num_workers = num_workers
        self.mutex = threading.Lock()
        self.no_output = no_output
        self._client_api = client_api
        self.key = datetime.datetime.now().strftime('%Y-%m-%d-%H-%M-%S')
        self.cache_mode = client_api.cache_state.enable_cache
        if self.cache_mode is None:
            self.cache_mode = True
        self.cache_chunk = client_api.cache_state.chunk_cache
        if self.cache_chunk is None:
            self.cache_chunk = CHUNK
        self._reports = None
        self._success = dict()
        self._status = dict()
        self._errors = dict()
        self._output = dict()
        self.cache_items = {'errors': self._errors,
                            'status': self._status,
                            'output': self._output}
        self._resource = resource
        self._print_error_logs = print_error_logs
        self._output_entity = output_entity

    @property
    def has_errors(self):
        """
        return True if errors has occurred False otherwise
        """
        return self.failure_count > 0

    def build_cache(self):
        if self.cache_mode:
            try:
                self._reports = {'errors': dl_cache.DiskCache('errors-' + self.key),
                                 'output': dl_cache.DiskCache('output-' + self.key),
                                 'status': dl_cache.DiskCache('status-' + self.key),
                                 }
                self._reports.get('status').add('success', 0)
                self._reports.get('status').add('failure', 0)
                self.clear_reporter()
            except:
                raise exceptions.PlatformException(
                    error='2001',
                    message='Failed to initialize cache handler. Please disable cache usage:  dl.cache_state.enable_cache = False')

    def construct_output(self, entity):
        """
        convert the json to his entity object
        """
        if self._output_entity and entity:
            return self._output_entity.from_json(client_api=self._client_api, _json=entity)
        return entity

    @property
    def output(self):
        """
        return a generator for all the outputs or the output it self if it is single
        """
        if self.no_output:
            return
        if self.cache_mode == 'diskcache':
            output = self._reports.get('output')
            for k in output.keys():
                for item in list(output.get(key=k).values()):
                    yield self.construct_output(item)

        current = [self.construct_output(output) for output in list(self._output.values()) if output is not None]
        for item in current:
            yield item

    @property
    def status_list(self):
        """
        return a list of all the status that get
        """
        output = dict()
        if self.cache_mode == 'diskcache':
            status_cache = self._reports.get('status')
            for key in status_cache.keys():
                if key not in ['failure', 'success']:
                    output.update(status_cache.get(key=key))
        output.update(self._status)
        return list(output.values())

    @property
    def num_workers(self):
        """
        number of the threads to work
        """
        return self._num_workers

    def upcount_num_workers(self):
        """
        increase the number of the threads to work
        """
        self._num_workers += 1

    @property
    def failure_count(self):
        """
        return how many actions fail
        """
        curr = len([suc for suc in (self._success.values()) if suc is False])
        if self.cache_mode == 'diskcache':
            return self._reports.get('status').get(key='failure') + curr
        else:
            return curr

    @property
    def success_count(self):
        """
        return how many actions success
        """
        curr = len([suc for suc in (self._success.values()) if suc is True])
        if self.cache_mode == 'diskcache':
            return self._reports.get('status').get(key='success') + curr
        else:
            return curr

    def status_count(self, status):
        """
        :param status: str of status to check it
        :return: how many times this status appear
        """
        status_list = self.status_list
        return list(status_list).count(status)

    def _write_to_disk(self):
        """
        the function write to the dick the outputs that get until the chunk amount
        """
        with self.mutex:
            if self._reports is None:
                self.build_cache()
            if len(self._success) > self.cache_chunk:
                status_cache = self._reports.get('status')
                num_true = sum(list(self._success.values()))
                status_cache.incr(key='failure',
                                  value=len(self._success) - num_true)
                status_cache.incr(key='success',
                                  value=num_true)
                self._success.clear()
            for name, cache_list in self.cache_items.items():
                if len(cache_list) > self.cache_chunk:
                    self._reports[name].push(cache_list)
                    cache_list.clear()

    def set_index(self, ref, error=None, status=None, success=None, output=None):
        """
        set the values that we get from the actions in the reporter
        """
        if self.mutex.locked():
            self.mutex.acquire()
            self.mutex.release()
        if error is not None:
            self._errors[ref] = error

        if status is not None:
            self._status[ref] = status

        if success is not None:
            self._success[ref] = success

        if success and output is not None:
            if not self.no_output:
                self._output[ref] = output

        if len(self._errors) > self.cache_chunk or \
                len(self._status) > self.cache_chunk or \
                len(self._output) > self.cache_chunk or \
                len(self._success) > self.cache_chunk:
            if self.cache_mode:
                self._write_to_disk()

    def generate_log_files(self):
        """
        build a log file that display the errors
        """
        if len(self._errors) > 0 and self.cache_mode == 'diskcache':
            # write from RAM to disk
            self._reports['errors'].push(self._errors)
            self._errors.clear()

        reports_dir = os.path.join(services.service_defaults.DATALOOP_PATH, 'reporters')
        if not os.path.exists(reports_dir):
            os.mkdir(reports_dir)
        log_filepath = os.path.join(reports_dir,
                                    "log_{}_{}.json".format(self._resource,
                                                            datetime.datetime.now(datetime.timezone.utc).strftime("%Y%m%d_%H%M%S%f"))) # Added %f to the format to include microseconds
        errors_json = dict()
        if self.cache_mode == 'diskcache':
            err_cache = self._reports['errors']
            for k in err_cache.keys():
                errors_json.update(err_cache.get(k))
        if self._errors:
            errors_json.update(self._errors)
        if self._print_error_logs:
            for key in errors_json:
                logger.warning("{}\n{}".format(key, errors_json[key]))
            return None
        else:
            with open(log_filepath, "w") as f:
                json.dump(errors_json, f, indent=2)
            return log_filepath

    def clear_reporter(self):
        """
        clear the file system from the outputs
        """
        import shutil
        date_now = datetime.datetime(year=int(self.key.split('-')[0]),
                                     month=int(self.key.split('-')[1]),
                                     day=int(self.key.split('-')[2]))
        cache_dir = os.path.dirname(self._reports['output'].cache_dir)
        cache_files_list = os.listdir(cache_dir)
        # remove all the cache files from the last day
        for filename in cache_files_list:
            try:
                date_file = datetime.datetime(year=int(filename.split('-')[1]),
                                              month=int(filename.split('-')[2]),
                                              day=int(filename.split('-')[3]))
            except:
                continue
            if date_file < date_now:
                try:
                    shutil.rmtree(cache_dir + '\\' + filename)
                except OSError:  # Windows wonkiness
                    pass



import attr
import logging
from .. import repositories, exceptions, entities

logger = logging.getLogger(name='dtlpy')


@attr.s
class Assignment(entities.BaseEntity):
    """
    Assignment object
    """

    # platform
    name = attr.ib()
    annotator = attr.ib()
    status = attr.ib()
    project_id = attr.ib()
    metadata = attr.ib(repr=False)
    id = attr.ib()
    url = attr.ib(repr=False)
    updated_at = attr.ib(repr=False)
    updated_by = attr.ib(repr=False)
    task_id = attr.ib(repr=False)
    dataset_id = attr.ib(repr=False)
    annotation_status = attr.ib(repr=False)
    item_status = attr.ib(repr=False)
    total_items = attr.ib()
    for_review = attr.ib()
    issues = attr.ib()
    # sdk

    _client_api = attr.ib(repr=False)
    _task = attr.ib(default=None, repr=False)
    _assignments = attr.ib(default=None, repr=False)
    _project = attr.ib(default=None, repr=False)
    _dataset = attr.ib(default=None, repr=False)
    _datasets = attr.ib(default=None, repr=False)

    @classmethod
    def from_json(cls, _json, client_api, project=None, task=None, dataset=None):
        if project is not None:
            if project.id != _json.get('projectId', None):
                logger.warning('Assignment has been fetched from a project that is not belong to it')
                project = None

        metadata = _json.get('metadata', dict())
        dataset_id = metadata.get('datasetId', None)
        task_id = metadata.get('taskId', None)
        if dataset_id is None:
            system_metadata = metadata.get('system', dict())
            dataset_id = system_metadata.get('datasetId', None)
        if task_id is None:
            system_metadata = metadata.get('system', dict())
            task_id = system_metadata.get('taskId', None)

        assignment = cls(
            name=_json.get('name', None),
            annotator=_json.get('annotator', None),
            status=_json.get('status', None),
            project_id=_json.get('projectId', None),
            task_id=task_id,
            dataset_id=dataset_id,
            metadata=metadata,
            url=_json.get('url', None),
            id=_json['id'],
            updated_by=_json.get('updatedBy', None),
            updated_at=_json.get('updatedAt', None),
            client_api=client_api,
            project=project,
            dataset=dataset,
            task=task,
            annotation_status=_json.get('annotationStatus', None),
            item_status=_json.get('itemStatus', None),
            total_items=_json.get('totalItems', None),
            for_review=_json.get('forReview', None),
            issues=_json.get('issues', None)
        )

        if dataset is not None:
            if assignment.dataset_id != dataset.id:
                logger.warning('Assignment has been fetched from a dataset that is not belong to it')
                assignment._dataset = None

        if task is not None:
            if assignment.task_id != task.id:
                logger.warning('Assignment has been fetched from a task that is not belong to it')
                assignment._task = None

        return assignment

    @property
    def platform_url(self):
        return self._client_api._get_resource_url("projects/{}/assignments/{}".format(self.project_id, self.id))

    @property
    def assignments(self):
        if self._assignments is None:
            self._assignments = repositories.Assignments(client_api=self._client_api,
                                                         project=self._project,
                                                         task=self._task,
                                                         dataset=self._dataset)
        assert isinstance(self._assignments, repositories.Assignments)
        return self._assignments

    @property
    def project(self):
        if self._project is None:
            self._project = repositories.Projects(client_api=self._client_api).get(project_id=self.project_id,
                                                                                   fetch=None)

        assert isinstance(self._project, entities.Project)
        return self._project

    @property
    def datasets(self):
        if self._datasets is None:
            self._datasets = repositories.Datasets(client_api=self._client_api, project=self._project)
        assert isinstance(self._datasets, repositories.Datasets)
        return self._datasets

    @property
    def dataset(self):
        if self._dataset is None:
            self._dataset = self.datasets.get(dataset_id=self.dataset_id)
        assert isinstance(self._dataset, entities.Dataset)
        return self._dataset

    @property
    def task(self):
        if self._task is None:
            self._task = self.project.tasks.get(task_id=self.task_id)
        assert isinstance(self._task, entities.Task)
        return self._task

    def to_json(self):
        """
        Returns platform _json format of object

        :return: platform json format of object
        :rtype: dict
        """
        _json = attr.asdict(self, filter=attr.filters.exclude(attr.fields(Assignment)._client_api,
                                                              attr.fields(Assignment)._project,
                                                              attr.fields(Assignment).project_id,
                                                              attr.fields(Assignment)._assignments,
                                                              attr.fields(Assignment)._dataset,
                                                              attr.fields(Assignment)._task,
                                                              attr.fields(Assignment).annotation_status,
                                                              attr.fields(Assignment).updated_at,
                                                              attr.fields(Assignment).updated_by,
                                                              attr.fields(Assignment).item_status,
                                                              attr.fields(Assignment).total_items,
                                                              attr.fields(Assignment).for_review,
                                                              attr.fields(Assignment).issues))

        _json['projectId'] = self.project_id
        return _json

    def update(self, system_metadata=False):
        """
        Update an assignment

        **Prerequisites**: You must be in the role of an *owner*, *developer*, or *annotation manager* who has been assigned as *owner* of the annotation task.

        :param bool system_metadata: True, if you want to change metadata system
        :return: Assignment object
        :rtype: dtlpy.entities.assignment.Assignment assignment

        **Example**:

        .. code-block:: python

            assignment = assignment.update(system_metadata=False)
        """
        return self.assignments.update(assignment=self, system_metadata=system_metadata)

    def open_in_web(self):
        """
        Open the assignment in web platform

        **Prerequisites**: You must be in the role of an *owner*, *developer*, or *annotation manager* who has been assigned as *owner* of the annotation task.

        :return:

        **Example**:

        .. code-block:: python

            assignment.open_in_web()
        """
        self._client_api._open_in_web(url=self.platform_url)

    def get_items(self, dataset=None, filters=None):
        """
        Get all the items in the assignment

        **Prerequisites**: You must be in the role of an *owner*, *developer*, or *annotation manager* who has been assigned as *owner* of the annotation task.

        :param dtlpy.entities.dataset.Dataset dataset: dataset object, the dataset that refer to the assignment
        :param dtlpy.entities.filters.Filters filters: Filters entity or a dictionary containing filters parameters
        :return: pages of the items
        :rtype: dtlpy.entities.paged_entities.PagedEntities

        **Example**:

        .. code-block:: python

            items = task.assignments.get_items()
        """
        if dataset is None:
            dataset = self.dataset

        return self.assignments.get_items(dataset=dataset, assignment=self, filters=filters)

    def reassign(self, assignee_id, wait=True):
        """
        Reassign an assignment

        **Prerequisites**: You must be in the role of an *owner*, *developer*, or *annotation manager* who has been assigned as *owner* of the annotation task.

        :param str assignee_id: the email of the user that want to assign the assignment
        :param bool wait: wait until reassign assignment finish
        :return: Assignment object
        :rtype: dtlpy.entities.assignment.Assignment

        **Example**:

        .. code-block:: python

            assignment = assignment.reassign(assignee_ids='annotator1@dataloop.ai')
        """
        return self.assignments.reassign(assignment=self,
                                         task=self._task,
                                         task_id=self.metadata['system'].get('taskId'),
                                         assignee_id=assignee_id,
                                         wait=wait)

    def redistribute(self, workload, wait=True):
        """
        Redistribute an assignment

        **Prerequisites**: You must be in the role of an *owner*, *developer*, or *annotation manager* who has been assigned as *owner* of the annotation task.

        :param dtlpy.entities.assignment.Workload workload: list of WorkloadUnit objects. Customize distribution (percentage) between the task assignees. For example: [dl.WorkloadUnit(annotator@hi.com, 80), dl.WorkloadUnit(annotator2@hi.com, 20)]
        :param bool wait: wait until redistribute assignment finish
        :return: Assignment object
        :rtype: dtlpy.entities.assignment.Assignment assignment

        **Example**:

        .. code-block:: python

            assignment = assignment.redistribute(workload=dl.Workload([dl.WorkloadUnit(assignee_id="annotator1@dataloop.ai", load=50),
                                                         dl.WorkloadUnit(assignee_id="annotator2@dataloop.ai", load=50)]))
        """
        return self.assignments.redistribute(assignment=self,
                                             task=self._task,
                                             task_id=self.metadata['system'].get('taskId'),
                                             workload=workload,
                                             wait=wait)

    def set_status(self, status: str, operation: str, item_id: str):
        """
        Set item status within assignment

        **Prerequisites**: You must be in the role of an *owner*, *developer*, or *annotation manager* who has been assigned as *owner* of the annotation task.

        :param str status: string the describes the status
        :param str operation: the status action need 'create' or 'delete'
        :param str item_id: item id that want to set his status
        :return: True id success
        :rtype: bool

        **Example**:

        .. code-block:: python

            success = assignment.set_status(status='complete',
                                    operation='created',
                                    item_id='item_id')
        """
        return self.assignments.set_status(status=status, operation=operation, item_id=item_id, assignment_id=self.id)


@attr.s
class WorkloadUnit:
    """
    WorkloadUnit object
    """

    # platform
    assignee_id = attr.ib(type=str)
    load = attr.ib(type=float, default=0)

    # noinspection PyUnusedLocal
    @load.validator
    def check_load(self, attribute, value):
        if value < 0 or value > 100:
            raise exceptions.PlatformException('400', 'Value must be a number between 0 to 100')

    @classmethod
    def from_json(cls, _json):
        return cls(
            assignee_id=_json.get('assigneeId', None),
            load=_json.get('load', None)
        )

    def to_json(self):
        return {
            'assigneeId': self.assignee_id,
            'load': self.load
        }


@attr.s
class Workload:
    """
    Workload object
    """

    # platform
    workload = attr.ib(type=list)

    def __iter__(self):
        for w_l_u in self.workload:
            yield w_l_u

    @workload.default
    def set_workload(self):
        workload = list()
        return workload

    @classmethod
    def from_json(cls, _json):
        return cls(
            workload=[WorkloadUnit.from_json(workload_unit) for workload_unit in _json]
        )

    @staticmethod
    def _loads_are_correct(loads):
        return round(sum(loads)) == 100

    @staticmethod
    def _get_loads(num_assignees):
        loads = [0 for _ in range(num_assignees)]
        index = 0
        for i in range(10000):
            loads[index] += 1
            if index < num_assignees - 1:
                index += 1
            else:
                index = 0
        loads = [l / 100 for l in loads]
        return loads

    def _redistribute(self):
        load = self._get_loads(num_assignees=len(self.workload))
        for i_w_l, w_l in self.workload:
            w_l.load = load

    @classmethod
    def generate(cls, assignee_ids, loads=None):
        """
        generate the loads for the given assignee
        :param assignee_ids:
        :param loads:
        """
        if not isinstance(assignee_ids, list):
            assignee_ids = [assignee_ids]

        if loads is None:
            div = len(assignee_ids)
            loads = [100 // div + (1 if x < 100 % div else 0) for x in range(div)]
        else:
            if not isinstance(loads, list):
                loads = [loads]
            if not Workload._loads_are_correct(loads=loads):
                raise exceptions.PlatformException('400', 'Loads must summarized to 100')

        if len(assignee_ids) != len(loads):
            raise exceptions.PlatformException('400', 'Assignee ids and loads must be of same length')

        return cls(
            workload=[WorkloadUnit(assignee_id=assignee_id, load=loads[i_assignee_id]) for i_assignee_id, assignee_id in
                      enumerate(assignee_ids)])

    def to_json(self):
        return [workload_unit.to_json() for workload_unit in self.workload]

    def add(self, assignee_id):
        """
        add a assignee

        :param assignee_id:
        """
        self.workload.append(WorkloadUnit(assignee_id=assignee_id))
        if not self._loads_are_correct(loads=[w_l.load for w_l in self.workload]):
            self._redistribute()
